# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, none, instance_of

from storops.vnx import converter
from storops.vnx.resource.lun import VNXLun, VNXLunList
from storops.vnx.resource.snap import VNXSnap
from storops_test.vnx.cli_mock import t_cli, patch_cli

__author__ = 'Cedric Zhuang'


class ConverterTest(TestCase):
    def test_id_to_lun_normal(self):
        lun = converter.id_to_lun('12')
        assert_that(lun, instance_of(VNXLun))
        assert_that(lun.get_id(lun), equal_to(12))

    def test_id_to_lun_na(self):
        lun = converter.id_to_lun('N/A')
        assert_that(lun, none())

    def test_id_to_lun_none(self):
        lun = converter.id_to_lun(None)
        assert_that(lun, none())

    def test_name_to_lun_na(self):
        lun = converter.name_to_lun('N/A')
        assert_that(lun, none())

    def test_name_to_lun_normal(self):
        lun = converter.name_to_lun('abc')
        assert_that(lun._name, equal_to('abc'))

    def name_to_snap(self):
        snap = converter.name_to_snap('abc')
        assert_that(snap, instance_of(VNXSnap))
        assert_that(snap._name, equal_to('abc'))

    def test_name_to_snap_na(self):
        snap = converter.name_to_snap('N/A')
        assert_that(snap, none())

    @patch_cli
    def test_ids_to_lun_list_normal(self):
        lun_list = converter.ids_to_lun_list('0,1,2')
        lun_list._cli = t_cli()
        assert_that(lun_list, instance_of(VNXLunList))
        assert_that(len(lun_list), equal_to(3))

    def test_ids_to_lun_list_empty(self):
        lun_list = converter.ids_to_lun_list('')
        assert_that(len(lun_list), equal_to(0))
        lun_list = converter.ids_to_lun_list(None)
        assert_that(len(lun_list), equal_to(0))

    @patch_cli
    def test_indices_to_disk_list_normal(self):
        value = '''Bus 0 Enclosure 0  Disk A0
                   Bus 0 Enclosure 0  Disk A4'''
        disks = converter.indices_to_disk_list(value)
        disks._cli = t_cli()
        assert_that(len(disks), equal_to(2))

    def test_indices_to_disk_list_empty(self):
        disks = converter.indices_to_disk_list('')
        assert_that(len(disks), equal_to(0))



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, none

from storops.vnx.resource.metric import VNXStats
from storops_test.vnx.cli_mock import t_cli, patch_cli

__author__ = 'Cedric Zhuang'


class VNXStatsTest(TestCase):
    stats = VNXStats.get(t_cli())

    @patch_cli
    def test_get_stats_status_disabled(self):
        assert_that(self.stats.is_enabled(), equal_to(False))

    @patch_cli(output='setstats_enabled.txt')
    def test_get_stats_status_enabled(self):
        assert_that(self.stats.is_enabled(), equal_to(True))

    @patch_cli
    def test_enable_stats(self):
        assert_that(self.stats.enable_stats(), none())

    @patch_cli
    def test_disable_stats(self):
        assert_that(self.stats.disable_stats(), none())



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

import ddt
import hamcrest as hc

import storops
from storops.unity.resource import import_session, remote_system, \
    replication_session, host
from storops_test.unity import rest_mock as mock

__author__ = 'Yong Huang'


def get_resource_config():
    resource_type = storops.ReplicationEndpointResourceTypeEnum.LUN
    return replication_session.UnityResourceConfig.to_embedded(
        pool_id='pool_1',
        is_thin_enabled=True,
        is_deduplication_enabled=True,
        is_compression_enabled=True,
        is_backup_only=True,
        size=1024,
        tiering_policy=storops.TieringPolicyEnum.AUTOTIER,
        request_id='request_1',
        src_id='src_1',
        name='resource_config_1',
        default_sp=storops.NodeEnum.SPA,
        replication_resource_type=resource_type
    )


@ddt.ddt
class UnityImportSessionTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        sess = import_session.UnityImportSession.get(cli=mock.t_rest(),
                                                     _id='import_65558')
        element_import_0 = sess.element_imports[0]
        hc.assert_that(sess.id, hc.equal_to('import_65558'))
        hc.assert_that(sess.state,
                       hc.equal_to(storops.ImportStateEnum.READY_TO_CUTOVER))
        hc.assert_that(sess.status,
                       hc.equal_to(
                           storops.ImportOpStatusEnum.READY_TO_CUTOVER))
        hc.assert_that(sess.operational_status[0],
                       hc.equal_to(
                           storops.ImportOpStatusEnum.READY_TO_CUTOVER))
        hc.assert_that(sess.type, hc.equal_to(storops.ImportTypeEnum.BLOCK))
        hc.assert_that(sess.resource_type,
                       hc.equal_to(storops.RemoteObjectTypeEnum.LUN))
        hc.assert_that(sess.target_resource_type,
                       hc.equal_to(storops.PoolConsumerTypeEnum.LUN))
        hc.assert_that(sess.instance_id, hc.equal_to('import_65558'))
        hc.assert_that(sess.name, hc.equal_to(
            'Import_APM00152904558_541_20190305062151.415258+000'))
        hc.assert_that(sess.health.value, hc.equal_to(storops.HealthEnum.OK))
        hc.assert_that(sess.src_resource.res_id, hc.equal_to('541'))
        hc.assert_that(sess.src_resource.name, hc.equal_to('lun-liangr'))
        hc.assert_that(sess.src_resource.system.id, hc.equal_to('RS_65557'))
        hc.assert_that(sess.target_resource.res_id, hc.equal_to(
            '60:06:01:60:74:E0:3A:00:80:15:7E:5C:EF:22:24:08'))
        hc.assert_that(sess.target_resource.name, hc.equal_to('lun-liangr-1'))
        hc.assert_that(sess.target_resource.system.id, hc.equal_to(
            'RS_65557'))
        hc.assert_that(len(sess.element_imports), hc.equal_to(1))
        hc.assert_that(element_import_0.src_id, hc.equal_to(
            'APM00152904558_60:06:01:60:12:90:38:00:F7:99:07:16:BE:DF:E7:11'))
        hc.assert_that(element_import_0.src_name, hc.equal_to('lun-liangr'))
        hc.assert_that(element_import_0.tgt_id, hc.equal_to('sv_2564'))
        hc.assert_that(element_import_0.tgt_name, hc.equal_to('lun-liangr-1'))
        hc.assert_that(sess.throttle, hc.equal_to(False))
        hc.assert_that(sess.remaining_data_size, hc.equal_to(0))
        hc.assert_that(sess.cutover_threshold, hc.equal_to(5))
        hc.assert_that(sess.remaining_data_percent, hc.equal_to(0))
        hc.assert_that(sess.estimated_cutover_window, hc.equal_to(0))
        hc.assert_that(sess.remote_system.id, hc.equal_to('RS_65557'))

    @mock.patch_rest
    def test_get_collection(self):
        sess_list = import_session.UnityImportSessionList.get(
            mock.t_rest())
        hc.assert_that(len(sess_list), hc.equal_to(3))
        hc.assert_that([sess.id for sess in sess_list],
                       hc.equal_to(
                           ['import_65558', 'import_65560', 'import_65562']))

    rc = get_resource_config()

    @mock.patch_rest
    @ddt.data(
        {'remote_sys_id': 'RS_65557',
         'src_resource_id': 'resource_1',
         'dst_resource_config': rc,
         'dst_resource_element_configs': [rc],
         'import_as_vmware_datastore': True,
         'throttle': False,
         'hosts':
             [host.UnityHost(_id='Host_27', cli=mock.t_rest())],
         'cutover_threshold': 80,
         'name': 'test_import_session_1'},
        {'remote_sys_id': 'RS_65557',
         'src_resource_id': 'resource_1',
         'dst_resource_config': rc,
         'dst_resource_element_configs': None,
         'import_as_vmware_datastore': None,
         'throttle': None,
         'hosts': None,
         'cutover_threshold': None,
         'name': None})
    @ddt.unpack
    def test_create_block_import(
            self, remote_sys_id, src_resource_id, dst_resource_config,
            dst_resource_element_configs=None,
            import_as_vmware_datastore=None,
            throttle=None, hosts=None, cutover_threshold=None, name=None):
        remote_sys = remote_system.UnityRemoteSystem(_id=remote_sys_id,
                                                     cli=mock.t_rest())
        src_resource = replication_session.UnityResourceInfo.to_embedded(
            res_id=src_resource_id, name=src_resource_id.upper(),
            system=remote_sys)
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.create_block_import(
            cli=mock.t_rest(),
            remote_system=remote_sys,
            src_resource=src_resource,
            dst_resource_config=dst_resource_config,
            dst_resource_element_configs=dst_resource_element_configs,
            import_as_vmware_datastore=import_as_vmware_datastore,
            throttle=throttle, hosts=hosts,
            cutover_threshold=cutover_threshold,
            name=name)
        hc.assert_that(resp.id, hc.equal_to('import_65564'))

    @mock.patch_rest
    @ddt.data(
        {'throttle': True,
         'name': 'import_session_1',
         'cutover_threshold': 80,
         'src_local_cifs_admin_username': 'user_1',
         'src_local_cifs_admin_passwd': 'passwd_1'},
        {'throttle': None,
         'name': None,
         'cutover_threshold': None,
         'src_local_cifs_admin_username': None,
         'src_local_cifs_admin_passwd': None})
    @ddt.unpack
    def test_modify(self, throttle=None, name=None, cutover_threshold=None,
                    src_local_cifs_admin_username=None,
                    src_local_cifs_admin_passwd=None):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.modify(
            throttle=throttle, name=name,
            cutover_threshold=cutover_threshold,
            src_local_cifs_admin_username=src_local_cifs_admin_username,
            src_local_cifs_admin_passwd=src_local_cifs_admin_passwd)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data(
        {'test_facility_code': 'test_facility_code_1'},
        {'test_facility_code': None})
    @ddt.unpack
    def test_pause(self, test_facility_code=None):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.pause(test_facility_code=test_facility_code)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    def test_resume(self):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.resume()
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data(
        {'cifs_server_info':
            import_session.UnityCifsServerInfo.to_embedded(
                name='cifs_server_1', description='CIFS Server 1',
                netbios_name='smb_server_1', domain='domain_1',
                workgroup='group_1', is_standalone=True,
                cifs_username='cifs_user_1'),
            'domain_username': 'ad_user_1',
            'domain_password': 'ad_passwd_1'},
        {'cifs_server_info': None,
         'domain_username': None,
         'domain_password': None})
    @ddt.unpack
    def test_cutover(self, cifs_server_info=None, domain_username=None,
                     domain_password=None):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.cutover(cifs_server_info=cifs_server_info,
                            domain_username=domain_username,
                            domain_password=domain_password)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    def test_commit(self):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.commit()
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data(
        {'domain_username': 'ad_user_1', 'domain_password': 'ad_passwd_1'},
        {'domain_username': None, 'domain_password': None})
    @ddt.unpack
    def test_cancel(self, domain_username=None, domain_password=None):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.cancel(domain_username=domain_username,
                           domain_password=domain_password)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    def test_delete(self):
        sess = import_session.UnityImportSession(_id='import_65558',
                                                 cli=mock.t_rest())
        resp = sess.delete()
        hc.assert_that(resp.is_ok(), hc.equal_to(True))



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import storops.unity.resource.nas_server
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class UnityFileDnsServer(UnityResource):
    @classmethod
    def create(cls, cli, nas_server, domain, ip_list):
        clz = storops.unity.resource.nas_server.UnityNasServer
        nas_server = clz.get(cli, nas_server)

        resp = cli.post(cls().resource_class,
                        nasServer=nas_server,
                        domain=domain,
                        addresses=ip_list)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)


class UnityFileDnsServerList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityFileDnsServer


class UnityFileDNSServerSourceParameters(UnityAttributeResource):
    pass



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.lib.common import text_var, yes_no_var
from storops.vnx.resource.cifs_share import CifsAccessControl

__author__ = 'Cedric Zhuang'


class NasCommand(object):
    @staticmethod
    def create_fs(fs_name, source_fs_name, pool_name, is_thin=False):
        cmd = ['/nas/bin/nas_fs']
        cmd += text_var('-name', fs_name)
        cmd += ['-type', 'uxfs', '-create']
        cmd.append('samesize={}'.format(source_fs_name))
        cmd.append('pool={}'.format(pool_name))
        cmd += ['storage=SINGLE', 'worm=off']
        cmd += yes_no_var('-thin', is_thin)
        cmd += ['-option', 'slice=y']
        return cmd

    @staticmethod
    def mount_fs(mover_name, fs_name, path=None):
        if path is None:
            path = '/{}'.format(fs_name)
        cmd = ['/nas/bin/server_mount', mover_name, '-option', 'ro']
        cmd += [fs_name, path]
        return cmd

    @staticmethod
    def copy_ckpt(snap_name, fs_name, connect_id, session_name=None):
        if session_name is None:
            session_name = '{}:{}'.format(fs_name, snap_name)
        cmd = ['/nas/bin/nas_copy']
        cmd += text_var('-name', session_name[0:63])
        cmd.append('-source')
        cmd += text_var('-ckpt', snap_name)
        cmd.append('-destination')
        cmd += text_var('-fs', fs_name)
        cmd += text_var('-interconnect', 'id={}'.format(connect_id))
        cmd += ['-overwrite_destination', '-full_copy']
        return cmd

    @staticmethod
    def umount(mover_name, name):
        cmd = text_var('/nas/bin/server_umount', mover_name)
        cmd += text_var('-perm', name)
        return cmd

    @staticmethod
    def mount(mover_name, name, rw='rw', path=None):
        if path is None:
            path = '/{}'.format(name)
        cmd = text_var('/nas/bin/server_mount', mover_name)
        cmd += text_var('rw', rw)
        cmd += [name, path]
        return cmd

    @staticmethod
    def delete_ckpt(ckpt_name):
        cmd = ['/nas/bin/nas_fs']
        cmd += text_var('-delete', ckpt_name)
        cmd.append('-Force')
        return cmd

    @staticmethod
    def nas_cel_list():
        return '/nas/bin/nas_cel -interconnect -l'.split()

    @staticmethod
    def get_dm_interfaces(name=None, is_vdm=True):
        cmd = ['/nas/bin/nas_server', '-i']
        if is_vdm:
            cmd.append('-vdm')

        if name is not None:
            cmd.append(name)
        else:
            cmd.append('-all')
        return cmd

    @staticmethod
    def attach_nfs_interface(if_name, vdm_name=None):
        cmd = ['/nas/bin/nas_server']
        if vdm_name is not None:
            cmd += text_var('-vdm', vdm_name)
        cmd += text_var('-attach', if_name)
        return cmd

    @staticmethod
    def detach_nfs_interface(if_name, vdm_name=None):
        cmd = ['/nas/bin/nas_server']
        if vdm_name is not None:
            cmd += text_var('-vdm', vdm_name)
        cmd += text_var('-detach', if_name)
        return cmd

    @staticmethod
    def _share_access_cmd_prefix(mover_name, share_name):
        cmd = text_var('/nas/bin/.server_config', mover_name)
        cmd.append('-v')
        cmd += text_var('sharesd', share_name)
        return cmd

    @classmethod
    def disable_cifs_share_access(cls, share_name, mover_name):
        cmd = cls._share_access_cmd_prefix(mover_name, share_name)
        cmd += ['set', 'noaccess']
        return cmd

    @classmethod
    def allow_cifs_share_access(cls, share_name, mover_name, user_name, domain,
                                access=CifsAccessControl.FULL):
        cmd = cls._share_access_cmd_prefix(mover_name, share_name)
        access_str = cls._access_str(access, domain, user_name)
        cmd += text_var('grant', access_str)
        return cmd

    @staticmethod
    def _access_str(access, domain, user_name):
        return '{}@{}={}'.format(user_name, domain, access)

    @classmethod
    def deny_cifs_share_access(cls, share_name, mover_name, user_name, domain,
                               access=CifsAccessControl.FULL):
        cmd = cls._share_access_cmd_prefix(mover_name, share_name)
        access_str = cls._access_str(access, domain, user_name)
        cmd += text_var('revoke', access_str)
        return cmd



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

import ddt
import hamcrest as hc

import storops
from storops.unity.resource import remote_system, health
from storops_test.unity import rest_mock as mock

__author__ = 'Ryan Liang'


@ddt.ddt
class UnityRemoteSystemTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        rs = remote_system.UnityRemoteSystem.get(cli=mock.t_rest(), _id='RS_0')
        hc.assert_that(rs.alt_management_address_list, hc.none())
        hc.assert_that(rs.connection_type,
                       hc.equal_to(storops.ReplicationCapabilityEnum.ASYNC))
        hc.assert_that(rs.health, hc.instance_of(health.UnityHealth))
        hc.assert_that(rs.health.value, hc.equal_to(storops.HealthEnum.OK))
        hc.assert_that(rs.id, hc.equal_to('RS_0'))
        hc.assert_that(rs.local_spa_interfaces,
                       hc.equal_to(['128.221.255.12']))
        hc.assert_that(rs.local_spb_interfaces,
                       hc.equal_to(['128.221.255.13']))
        hc.assert_that(rs.management_address, hc.equal_to('10.245.101.39'))
        hc.assert_that(rs.model, hc.equal_to('Unity 500'))
        hc.assert_that(rs.name, hc.equal_to('FNM00150600267'))
        hc.assert_that(rs.remote_spa_interfaces,
                       hc.equal_to(['128.221.255.12']))
        hc.assert_that(rs.remote_spb_interfaces,
                       hc.equal_to(['128.221.255.13']))
        hc.assert_that(rs.serial_number, hc.equal_to('FNM00150600267'))
        hc.assert_that(rs.sync_fc_ports, hc.equal_to(['spa_fc4', 'spb_fc4']))
        hc.assert_that(rs.username, hc.equal_to('admin'))

    @mock.patch_rest
    def test_get_collection(self):
        rs_list = remote_system.UnityRemoteSystemList.get(mock.t_rest())
        hc.assert_that(len(rs_list), hc.equal_to(2))
        hc.assert_that([rs.id for rs in rs_list],
                       hc.equal_to(['RS_0', 'RS_4']))

    @mock.patch_rest
    @ddt.data({'management_address': '1.1.1.1',
               'local_username': None, 'local_password': None,
               'remote_username': None, 'remote_password': None,
               'connection_type': None},
              {'management_address': '1.1.1.1',
               'local_username': 'admin', 'local_password': 'password',
               'remote_username': 'admin', 'remote_password': 'password',
               'connection_type': storops.ReplicationCapabilityEnum.BOTH})
    @ddt.unpack
    def test_create(self, management_address, local_username, local_password,
                    remote_username, remote_password, connection_type):
        rs = remote_system.UnityRemoteSystem.create(
            mock.t_rest(), management_address,
            local_username=local_username, local_password=local_password,
            remote_username=remote_username, remote_password=remote_password,
            connection_type=connection_type)
        hc.assert_that(rs.id, hc.equal_to('RS_4'))

    @mock.patch_rest
    @ddt.data({'management_address': '1.1.1.1',
               'username': 'admin', 'password': 'password',
               'connection_type': storops.ReplicationCapabilityEnum.BOTH})
    @ddt.unpack
    def test_modify(self, management_address, username, password,
                    connection_type):
        rs = remote_system.UnityRemoteSystem.get(cli=mock.t_rest(), _id='RS_0')
        resp = rs.modify(management_address=management_address,
                         username=username, password=password,
                         connection_type=connection_type)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data({'connection_type': storops.ReplicationCapabilityEnum.BOTH})
    @ddt.unpack
    def test_verify(self, connection_type):
        rs = remote_system.UnityRemoteSystem.get(cli=mock.t_rest(), _id='RS_0')
        resp = rs.verify(connection_type=connection_type)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import equal_to, assert_that, raises

from storops import VNXUserRoleEnum, VNXUserScopeEnum
from storops.exception import VNXUserNameInUseError, VNXUserNotFoundError
from storops_test.vnx.cli_mock import t_cli, patch_cli

from storops.vnx.resource.security import VNXBlockUser

__author__ = 'Cedric Zhuang'


class VNXSecurityTest(TestCase):
    @patch_cli
    def test_get_existed_user(self):
        user = VNXBlockUser.get(name='a', cli=t_cli())
        assert_that(user.name, equal_to('a'))
        assert_that(user.role, equal_to(VNXUserRoleEnum.ADMIN))
        assert_that(user.scope, equal_to(VNXUserScopeEnum.GLOBAL))
        assert_that(user.type, equal_to('user'))

    @patch_cli
    def test_get_all_users(self):
        users = VNXBlockUser.get(cli=t_cli())
        assert_that(len(users), equal_to(2))

    @patch_cli
    def test_get_user_not_found(self):
        user = VNXBlockUser(name='c', cli=t_cli())
        assert_that(user.existed, equal_to(False))

    @patch_cli
    def test_create_user_success(self):
        user = VNXBlockUser.create(t_cli(), 'b', 'b')
        assert_that(user.name, equal_to('b'))
        assert_that(user.role, equal_to(VNXUserRoleEnum.ADMIN))

    @patch_cli
    def test_create_user_existed(self):
        def f():
            VNXBlockUser.create(t_cli(), 'b', 'b',
                                role=VNXUserRoleEnum.OPERATOR)

        assert_that(f, raises(VNXUserNameInUseError, 'failed'))

    @patch_cli
    def test_delete_user_success(self):
        user = VNXBlockUser.get(name='b', cli=t_cli())
        # no exception
        user.delete()

    @patch_cli
    def test_delete_user_not_found(self):
        def f():
            user = VNXBlockUser('c', cli=t_cli())
            user.delete()

        assert_that(f, raises(VNXUserNotFoundError, 'not exist'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.vnx.resource import VNXCliResourceList
from storops.vnx.resource import VNXCliResource

__author__ = 'Cedric Zhuang'


class VNXNduList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXNdu

    def _get_raw_resource(self):
        return self._cli.get_ndu(poll=self.poll)

    def _check_package(self, name):
        for ndu in self:
            if ndu.name == name:
                ret = VNXNdu.is_enabled(ndu)
                break
        else:
            ret = False
        return ret

    def is_dedup_enabled(self):
        return self._check_package(VNXNdu.DEDUP)

    def is_compression_enabled(self):
        return self._check_package(VNXNdu.COMPRESSION)

    def is_auto_tiering_enabled(self):
        return self._check_package(VNXNdu.AUTO_TIERING)

    def is_mirror_view_async_enabled(self):
        return self._check_package(VNXNdu.MIRROR_VIEW_ASYNC)

    def is_mirror_view_sync_enabled(self):
        return self._check_package(VNXNdu.MIRROR_VIEW_SYNC)

    def is_mirror_view_enabled(self):
        return (self.is_mirror_view_async_enabled() and
                self.is_mirror_view_sync_enabled())

    def is_sancopy_enabled(self):
        return self._check_package(VNXNdu.SANCOPY)

    def is_thin_enabled(self):
        return self._check_package(VNXNdu.THIN)

    def is_snap_enabled(self):
        return self._check_package(VNXNdu.SNAP)

    def is_fast_cache_enabled(self):
        return self._check_package(VNXNdu.FAST_CACHE)


class VNXNdu(VNXCliResource):
    DEDUP = '-Deduplication'
    COMPRESSION = '-Compression'
    AUTO_TIERING = '-FAST'
    MIRROR_VIEW_ASYNC = '-MirrorView/A'
    MIRROR_VIEW_SYNC = '-MirrorView/S'
    SANCOPY = '-SANCopy'
    THIN = '-ThinProvisioning'
    SNAP = '-VNXSnapshots'
    FAST_CACHE = '-FASTCache'

    def __init__(self, name=None, cli=None):
        super(VNXNdu, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_ndu(name=self._name, poll=self.poll)

    @staticmethod
    def is_enabled(ndu):
        return ndu.existed and ndu.active_state and not ndu.commit_required

    @classmethod
    def _check_package(cls, cli, name):
        ndu = VNXNdu(name, cli)
        ndu.with_no_poll()
        return cls.is_enabled(ndu)

    @classmethod
    def is_dedup_enabled(cls, cli):
        return cls._check_package(cli, cls.DEDUP)

    @classmethod
    def is_compression_enabled(cls, cli):
        return cls._check_package(cli, cls.COMPRESSION)

    @classmethod
    def is_auto_tiering_enabled(cls, cli):
        return cls._check_package(cli, cls.AUTO_TIERING)

    @classmethod
    def is_mirror_view_async_enabled(cls, cli):
        return cls._check_package(cli, cls.MIRROR_VIEW_ASYNC)

    @classmethod
    def is_mirror_view_sync_enabled(cls, cli):
        return cls._check_package(cli, cls.MIRROR_VIEW_SYNC)

    @classmethod
    def is_mirror_view_enabled(cls, cli):
        return (cls.is_mirror_view_async_enabled(
            cli) and cls.is_mirror_view_sync_enabled(cli))

    @classmethod
    def is_sancopy_enabled(cls, cli):
        return cls._check_package(cli, cls.SANCOPY)

    @classmethod
    def is_thin_enabled(cls, cli):
        return cls._check_package(cli, cls.THIN)

    @classmethod
    def is_snap_enabled(cls, cli):
        return cls._check_package(cli, cls.SNAP)

    @classmethod
    def is_fast_cache_enabled(cls, cli):
        return cls._check_package(cli, cls.FAST_CACHE)

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXNduList(cli)
        else:
            ret = VNXNdu(name, cli)
        return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from datetime import datetime, timedelta

import storops.exception as ex
from storops.lib.common import JsonPrinter, clear_instance_cache

__author__ = 'Cedric Zhuang'


class Resource(JsonPrinter):
    def __init__(self):
        self._parsed_resource = None
        super(Resource, self).__init__()

    def shadow_copy(self):
        """ Return a copy of the resource with same raw data

        :return: copy of the resource
        """
        ret = self.__class__()
        if not self._is_updated():
            # before copy, make sure source is updated.
            self.update()
        ret._parsed_resource = self._parsed_resource
        return ret

    def _get_name(self):
        if hasattr(self, '_name') and self._name is not None:
            name = self._name
        else:
            name = self.name
        return name

    @classmethod
    def _get_parser(cls):
        raise NotImplementedError('_get_parser not implemented.')

    @property
    def parsed_resource(self):
        return self._parsed_resource

    @clear_instance_cache
    def update(self, data=None):
        if data is None:
            data = self._get_raw_resource()

        self._parsed_resource = self._parse_raw(data)
        self.update_name_if_exists()
        return self

    def update_name_if_exists(self):
        if hasattr(self, '_name') and self._name is None:
            setattr(self, '_name', self._get_value_by_key("name"))

    def get_index(self):
        parser = self._get_parser()
        index_desc = parser.index_property
        if index_desc is not None:
            ret = getattr(self, index_desc.key)
        else:
            raise ex.NoIndexException('{} does not have index.'
                                      .format(self.__class__.__name__))
        return ret

    @property
    def existed(self):
        try:
            ret = self.get_index() is not None
        except ex.NoIndexException:
            # no index, check if any of the property is available
            prop = self._get_first_not_none_prop()
            ret = prop is not None
        return ret

    def _get_first_not_none_prop(self):
        ret = None
        prop_desc_list = self._get_parser().properties
        for prop_desc in prop_desc_list:
            ret = getattr(self, prop_desc.key)
            if ret is not None:
                break
        return ret

    def _parse_raw(self, data):
        return self._get_parser().parse(data)

    @classmethod
    def parse(cls, output):
        obj = cls()
        data = cls._get_parser().parse_single(output)
        return obj.update(data)

    @classmethod
    def parse_all(cls, output):
        ret = []
        for data in cls._get_parser().parse_all(output):
            obj = cls()
            ret.append(obj.update(data))
        return ret

    def property_names(self):
        parser = self._get_parser()
        if parser is not None:
            ret = parser.property_names
        else:
            ret = []
        return ret

    @classmethod
    def get_property_label(cls, key):
        parser = cls._get_parser()
        if parser is not None:
            ret = parser.get_property_label(key)
        else:
            ret = None
        return ret

    @property
    def system_version(self):
        # Return the version of backing system
        return None

    def _get_properties(self, dec=0):
        props = {'hash': self.__hash__()}

        if dec >= 0:
            prop_names = list(self.property_names())
            prop_names.append('existed')
            for name in prop_names:
                try:
                    value = getattr(self, name)
                    if isinstance(value, JsonPrinter):
                        value = value.get_dict_repr(dec - 1)
                    elif isinstance(value, (datetime, timedelta)):
                        value = str(value)
                    elif isinstance(value, (tuple, list, set)):
                        value = [v.get_dict_repr(dec - 1)
                                 if isinstance(v, JsonPrinter) else v
                                 for v in value]
                    props[name] = value
                except AttributeError:
                    # skip not available attributes
                    continue
        return props

    def _get_raw_resource(self):
        """get raw input of this resource

        Get the raw input of this resource.
        The input could be retrieved from multiple interface like
        CLI or CIM."""
        return ''

    def __getattr__(self, item):
        if not item.startswith('_'):
            ret = self._get_property_from_raw(item)
        else:
            raise AttributeError(item)
        return ret

    def _is_updated(self):
        return self._parsed_resource is not None

    def get_preloaded_prop_keys(self):
        return []

    def _get_property_from_raw(self, item):
        is_preloaded_prop = item in self.get_preloaded_prop_keys()
        if not self._is_updated() and not is_preloaded_prop:
            self.update()

        if item in self.property_names():
            ret = self._get_value_by_key(item)
        else:
            raise AttributeError(
                "'{}' does not contain attribute '{}'".format(
                    self.__class__.__name__, item))
        return ret

    def _get_value_by_key(self, item):
        return self._parsed_resource.get(item, None)


class ResourceList(Resource):
    def __init__(self):
        super(ResourceList, self).__init__()
        self._list = None

    def shadow_copy(self, *args, **kwargs):
        ret = super(ResourceList, self).shadow_copy()
        ret._list = self._list
        ret.set_filter(*args, **kwargs)
        return ret

    def set_filter(self, *args, **kwargs):
        self._set_filter(*args, **kwargs)
        self._apply_filter()

    def _set_filter(self, *args, **kwargs):
        # implemented by child classes if needed
        pass

    @classmethod
    def get_rsc_clz_list(cls, rsc_list_collection):
        return [l.get_resource_class() for l in rsc_list_collection]

    @clear_instance_cache
    def update(self, data=None):
        self._list = []
        if data is None:
            data = self._get_raw_resource()

        if data is not None and isinstance(data, dict):
            parsed_list = data
        else:
            parsed_list = self._parse_raw(data)

        for i in parsed_list:
            item = self._get_resource_instance()
            item.update(i)
            if self._filter(item):
                self._list.append(item)
        return self

    def _apply_filter(self):
        result = []
        for item in self:
            if self._filter(item):
                result.append(item)
        self._list = result

    def _filter(self, _):
        return True

    def _parse_raw(self, data):
        return self._get_parser().parse_all(data)

    @classmethod
    def _get_parser(cls):
        raise NotImplementedError('_get_parser not implemented.')

    @property
    def list(self):
        if self._list is None:
            self.update()
        return self._list

    @classmethod
    def get_resource_class(cls):
        raise NotImplementedError(
            'should return the class ref of the resource in the list.')

    def get_dict_repr(self, dec=1):
        items = [item.get_dict_repr(dec - 1) for item in self.list]
        return {self.__class__.__name__: items}

    def get_preloaded_prop_keys(self):
        return []

    def __len__(self):
        return len(self.list)

    def __iter__(self):
        return iter(self.list)

    def __getitem__(self, item):
        return self.list[item]

    def _get_resource_instance(self):
        return self.get_resource_class()()

    def __getattr__(self, v):
        s = self._get_resource_instance()
        ret = None
        if not v.startswith('_'):
            if v in dir(s):
                ret = self.get_member_attr_list(v)
            elif hasattr(s, 'property_names') and v in s.property_names():
                ret = self.get_member_attr_list(v)

        if ret is None:
            raise AttributeError(
                '{} do not has attribute {}.'.format(
                    self.get_resource_class().__name__, v))
        return ret

    def get_member_attr_list(self, v):
        return [getattr(i, v) for i in self]

    def _is_updated(self):
        return self._list is not None

    def __add__(self, other):
        if not isinstance(other, ResourceList):
            raise TypeError('Unsupported type {}'.format(type(other)))
        return self.list + other.list

    def append(self, item):
        if not isinstance(item, self.get_resource_class()):
            raise TypeError('Unsupported type {}'.format(type(item)))
        return self.list.append(item)


class ResourceListCollection(object):
    def __init__(self, init_list=None):
        if init_list:
            self._items = {rsc_list.get_resource_class(): rsc_list
                           for rsc_list in init_list}
        else:
            self._items = {}
        self.timestamp = datetime.now()

    def add_rsc_list(self, rsc_list):
        if not hasattr(rsc_list, 'get_resource_class'):
            raise ValueError('expect a ResourceList.')
        self._items[rsc_list.get_resource_class()] = rsc_list

    def get_rsc(self, obj):
        ret = None
        rsc_list = self.get_rsc_list(type(obj))
        if rsc_list:
            try:
                ret = rsc_list.get(obj)
            except TypeError:
                raise TypeError('"get(id)" method not found in ResourceList.')
        return ret

    def get_rsc_list(self, rsc_clz):
        return self._items.get(rsc_clz)

    def get_rsc_list_collection(self):
        return self._items.values()

    def get_rsc_clz_list(self):
        return self._items.keys()

    def __len__(self):
        return len(self._items)

    def update(self):
        for rsc_list in self.get_rsc_list_collection():
            rsc_list.update()
        self.timestamp = datetime.now()
        return self

    def delta_seconds(self, other):
        return (self.timestamp - other.timestamp).total_seconds()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import six
from hamcrest import assert_that, equal_to, none, only_contains

from storops.lib.parser import PropMapper, PropDescriptor, OutputParser

__author__ = 'Cedric Zhuang'


class PropMapperTest(TestCase):
    def test_camel_case_to_under_score(self):
        test_data = {
            'AbcDef': 'ABC_DEF',
            'abc def': 'ABC_DEF',
            'abc:': 'ABC',
            'SPAWrites': 'SPA_WRITES',
            'Is Thin LUN': 'IS_THIN_LUN',
            'TestCIMElement': 'TEST_CIM_ELEMENT'
        }

        for (k, v) in six.iteritems(test_data):
            assert_that(PropMapper.camel_case_to_under_score(k).upper(),
                        equal_to(v))

    def test_camel_case_to_under_score_with_delimiter(self):
        test_data = {
            'AbcDef': 'ABC.DEF',
            'abc def': 'ABC.DEF',
            'Is Thin LUN': 'IS.THIN.LUN',
            'abc:': 'ABC',
            'SPAWrites': 'SPA.WRITES',
            'TestCIMElement': 'TEST.CIM.ELEMENT',
            'VALUE_ARRAY': 'VALUE.ARRAY',
            'LUNs': 'LUNS',
            'Source LUN(s)': 'SOURCE.LUNS',
            'Capacity (GBs)': 'CAPACITY.GBS'
        }

        for (k, v) in six.iteritems(test_data):
            assert_that(PropMapper.camel_case_to_under_score(k, '.').upper(),
                        equal_to(v))


A = PropDescriptor('-a', 'Prop A (name):', 'prop_a')
B = PropDescriptor('-b', 'Prop B:')
C = PropDescriptor('-c', 'Prop C:')
ID = PropDescriptor(None, 'ID:', is_index=True)


class DemoParser(OutputParser):
    def __init__(self):
        super(DemoParser, self).__init__()
        self.add_property(A, B, C, ID)


class OutputParserTest(TestCase):
    def test_sequence(self):
        parser = DemoParser()
        index = parser.index_property
        assert_that(index.sequence, equal_to(3))

    def test_property_key(self):
        parser = DemoParser()
        assert_that(parser.has_property_key('prop_a'), equal_to(True))
        assert_that(parser.has_property_key('prop_b'), equal_to(True))
        assert_that(parser.has_property_key('D'), equal_to(False))

    def test_get_property_label_found(self):
        parser = DemoParser()
        assert_that(parser.get_property_label('prop_a'),
                    equal_to('Prop A (name):'))

    def test_get_property_label_not_found(self):
        parser = DemoParser()
        assert_that(parser.get_property_label('na'), none())

    def test_get_property_key_found(self):
        parser = DemoParser()
        assert_that(parser.get_property_key('Prop A (name):'),
                    equal_to('prop_a'))

    def test_get_property_key_not_found(self):
        parser = DemoParser()
        assert_that(parser.get_property_key('Prop B (name):'), none())

    def test_property_names(self):
        parser = DemoParser()
        assert_that(parser.property_names,
                    only_contains('id', 'prop_a', 'prop_b', 'prop_c'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import functools
import logging
import os
import re
from datetime import timedelta

import six
from mock import patch

from storops import VNXSPEnum
from storops.lib.common import cache, allow_omit_parentheses
from storops.lib.resource import ResourceListCollection
from storops.vnx.block_cli import CliClient
from storops.vnx.resource.disk import VNXDiskList
from storops.vnx.resource.lun import VNXLunList
from storops.vnx.resource.port import VNXSPPortList
from storops.vnx.resource.system import VNXSystem
from storops.vnx.resource.vnx_domain import VNXStorageProcessor, \
    VNXStorageProcessorList
from storops_test.utils import ConnectorMock

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class MockCli(ConnectorMock):
    base_folder = os.path.join('vnx', 'testdata', 'block_output')

    def get_folder(self, inputs):
        return self.base_folder

    def mock_execute(self, params, *args, **_):
        return self.get_mock_output(params)

    escaped_pattern = re.compile(r"[\\/:]")

    flags_to_delete = {
        '-t': 2
    }

    @classmethod
    def get_filename(cls, params):
        def delete_flag(arr, flag, flag_length=1):
            if flag in arr:
                i = arr.index(flag)
                if i > 0:
                    arr = arr[0:i] + arr[i + flag_length:]
                else:
                    arr = arr[flag_length:]
            return arr

        def delete_cli_binary(p):
            return p[1:]

        def delete_confidential(p):
            if p[0] == '-h':
                p = p[2:]
            if p[0] == '-user':
                p = p[2:]
            if p[0] == '-password':
                p = p[2:]
            if p[0] == '-scope':
                p = p[2:]
            return p

        params = delete_cli_binary(params)
        params = delete_confidential(params)

        for k, v in cls.flags_to_delete.items():
            params = delete_flag(params, k, v)
        name = '_'.join(map(six.text_type, params))
        name = re.sub(cls.escaped_pattern, '_', name)
        if len(name) >= 200:
            name = name[:43] + '____' + name[-43:]
        return '{}.txt'.format(name)


@allow_omit_parentheses
def patch_cli(output=None, mock_map=None):
    cli = MockCli(output, mock_map)

    def decorator(func):
        @functools.wraps(func)
        @patch(target='storops.vnx.navi_command.'
                      'NaviCommand.execute_naviseccli',
               new=cli.mock_execute)
        def func_wrapper(*args):
            return func(*args)

        return func_wrapper

    return decorator


def extract_command(func):
    """patch `CliClient.execute` method for unittest

    Patch `CliClient.execute` to return the parameters.
    The command line parameter can be verified by unittest.

    :param func: test function to wrap
    """

    def mock(_, commands, ip=None):
        if ip is None:
            pre = ''
        else:
            pre = '[{}] '.format(ip)
        return pre + ' '.join(map(six.text_type, commands))

    @functools.wraps(func)
    @patch(target='storops.vnx.block_cli.CliClient.execute', new=mock)
    @patch(target='storops.vnx.block_cli.CliClient.execute_dual', new=mock)
    def func_wrapper(self):
        return func(self)

    return func_wrapper


@cache
@patch_cli
def t_cli(version=None):
    """ get the test cli client

    :param with_stats: if true, returns a client with stats enabled.
    :param version: system version
    :return: test cli client
    """
    c = CliClient('10.244.212.182', heartbeat_interval=0)
    c.set_system_version(version)
    c.get_agent()
    prev = ResourceListCollection([get_sp_list_t0(c),
                                   get_lun_list_t0(c),
                                   get_disk_list_t0(c),
                                   get_port_list_t0(c)])
    curr = ResourceListCollection([get_sp_list_t1(c),
                                   get_lun_list_t1(c),
                                   get_disk_list_t1(c),
                                   get_port_list_t1(c)])
    curr.timestamp += timedelta(seconds=60)
    c.add_metric_record(prev)
    c.add_metric_record(curr)
    return c


@cache
def t_vnx():
    """ get the test vnx instance

    :return: test vnx instance
    """
    return VNXSystem('10.244.211.30', heartbeat_interval=0,
                     file_username='nasadmin',
                     file_password='nasadmin')


@patch_cli
def get_sp_list_t0(cli):
    sp = VNXStorageProcessor(cli, VNXSPEnum.SP_A, '10.244.211.30')
    sp.with_no_poll()
    ret = VNXStorageProcessorList(sp)
    ret.update()
    return ret


@patch_cli
def get_sp_list_t1(cli):
    sp = VNXStorageProcessor(cli, VNXSPEnum.SP_A, '10.244.211.30')
    ret = VNXStorageProcessorList(sp)
    ret.update()
    return ret


@patch_cli(output='lun_-list_-all_t0.txt')
def get_lun_list_t0(cli):
    return VNXLunList(cli=cli).update()


@patch_cli(output='getdisk_t0.txt')
def get_disk_list_t0(cli):
    return VNXDiskList(cli=cli).update()


@patch_cli(output='lun_-list_-all_t1.txt')
def get_lun_list_t1(cli):
    return VNXLunList(cli=cli).update()


@patch_cli(output='getdisk_t1.txt')
def get_disk_list_t1(cli):
    return VNXDiskList(cli=cli).update()


@patch_cli
def get_port_list_t0(cli):
    return VNXSPPortList(cli=cli).update()


@patch_cli(output='port_-list_-sp_-all_t1.txt')
def get_port_list_t1(cli):
    return VNXSPPortList(cli=cli).update()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import six

from storops.vnx.enums import VNXEnum
from storops.vnx.resource.mover import VNXMover
from storops.vnx.resource import VNXResource, VNXCliResourceList
from storops.vnx.resource.vdm import VNXVdm

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class CifsAccessControl(VNXEnum):
    FULL = 'fullcontrol'
    READ = 'read'


class VNXCifsShareList(VNXCliResourceList):
    def __init__(self, cli=None, server_name=None, share_name=None,
                 mover=None):
        super(VNXCifsShareList, self).__init__(cli=cli)
        self._server_name = server_name
        self._share_name = share_name
        self._mover = mover

    @classmethod
    def get_resource_class(cls):
        return VNXCifsShare

    def _get_raw_resource(self):
        if self._mover is not None:
            mover_id = self._mover.get_mover_id()
            is_vdm = self._mover.is_vdm
        else:
            mover_id = None
            is_vdm = None

        return self._cli.get_cifs_share(
            server_name=self._server_name,
            share_name=self._share_name,
            mover_id=mover_id,
            is_vdm=is_vdm)


class VNXCifsShare(VNXResource):
    def __init__(self, name=None, mover=None, cli=None):
        super(VNXCifsShare, self).__init__()
        self._name = name
        self._mover = mover
        self._cli = cli

    def _get_raw_resource(self):
        if self._mover is None:
            raise ValueError('mover is not specified for this share.')

        return self._cli.get_cifs_share(
            share_name=self._name,
            server_name=None,
            mover_id=self._mover.get_mover_id(),
            is_vdm=self._mover.is_vdm)

    @staticmethod
    def get(cli, name=None, mover=None, server_name=None):
        if name is not None and mover is not None:
            ret = VNXCifsShare(name, mover, cli)
        else:
            ret = VNXCifsShareList(
                cli, server_name=server_name, share_name=name, mover=mover)
        return ret

    @staticmethod
    def create(cli, fs, server_name, mover, path=None):
        mover_id = mover.get_mover_id()
        is_vdm = mover.is_vdm
        if not isinstance(fs, six.string_types):
            name = fs.get_name()
        else:
            name = fs

        resp = cli.create_cifs_share(name, server_name, mover_id, is_vdm, path)
        resp.raise_if_err()
        return VNXCifsShare(name=name, mover=mover, cli=cli)

    @property
    def mover(self):
        if self._mover is not None:
            ret = self._mover
        elif self.is_vdm:
            ret = VNXVdm(vdm_id=self.mover_id, cli=self._cli)
        else:
            ret = VNXMover(mover_id=self.mover_id, cli=self._cli)
        return ret

    def delete(self, *server_names):
        if not server_names:
            server_names = self.cifs_server_names
        resp = self._cli.delete_cifs_share(
            self._get_name(), self.mover.get_mover_id(), server_names,
            self.mover.is_vdm)
        resp.raise_if_err()
        return resp

    def disable_share_access(self):
        self._cli.disable_cifs_share_access(self._get_name(),
                                            self.mover._get_name())

    def allow_share_access(self, user_name, domain,
                           access=CifsAccessControl.FULL):
        self._cli.allow_cifs_share_access(
            self.mover._get_name(), self._get_name(), user_name, domain,
            access)

    def deny_share_access(self, user_name, domain,
                          access=CifsAccessControl.FULL):
        self._cli.deny_cifs_share_access(
            self.mover._get_name(), self._get_name(), user_name, domain,
            access)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from storops.vnx.resource.mover import VNXMoverRefList
from storops.vnx.resource import VNXCliResourceList, VNXResource

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class CifsDomain(object):
    def __init__(self, name, comp_name=None, user=None, password=None):
        self.name = name
        self.comp_name = comp_name
        self.user = user
        self.password = password


class VNXCifsServerList(VNXCliResourceList):
    def __init__(self, cli=None, mover_id=None, is_vdm=False):
        super(VNXCifsServerList, self).__init__(cli=cli)
        self.mover_id = mover_id
        self.is_vdm = is_vdm

    @classmethod
    def get_resource_class(cls):
        return VNXCifsServer

    def _get_raw_resource(self):
        return self._cli.get_cifs_server(mover_id=self.mover_id,
                                         is_vdm=self.is_vdm)


class VNXCifsServer(VNXResource):
    def __init__(self, name=None, cli=None):
        super(VNXCifsServer, self).__init__()
        self._name = name
        self._cli = cli

    def _get_raw_resource(self):
        return self._cli.get_cifs_server(name=self._name)

    @staticmethod
    def get(cli, name=None, mover_id=None, is_vdm=False):
        if name is not None:
            ret = VNXCifsServer(name=name, cli=cli)
        else:
            ret = VNXCifsServerList(cli=cli, mover_id=mover_id, is_vdm=is_vdm)
        return ret

    @staticmethod
    def create(cli, name, mover_id=None, is_vdm=False,
               workgroup=None, domain=None,
               interfaces=None, alias_name=None,
               local_admin_password=None):
        # default to first physical data mover
        if mover_id is None:
            movers = VNXMoverRefList(cli=cli)
            if not movers:
                raise ValueError('no data mover available.')
            mover_id = movers[0].mover_id
            is_vdm = False

        resp = cli.create_cifs_server(
            name=name, mover_id=mover_id, is_vdm=is_vdm,
            workgroup=workgroup, domain=domain,
            ip_list=interfaces, alias_name=alias_name,
            local_admin_password=local_admin_password)
        resp.raise_if_err()
        return VNXCifsServer(name=name, cli=cli)

    def delete(self, mover_id=None, is_vdm=False):
        if mover_id is None:
            mover_id = self.mover_id
            is_vdm = self.is_vdm
        resp = self._cli.delete_cifs_server(self._get_name(), mover_id, is_vdm)
        resp.raise_if_err()
        return resp

    def modify(self, name, mover_id=None, is_vdm=True,
               join_domain=False, username=None, password=None):
        if mover_id is None:
            mover_id = self.mover_id
            is_vdm = self.is_vdm
        resp = self._cli.modify_domain_cifs_server(
            name, mover_id, is_vdm, join_domain, username, password)
        resp.raise_if_err()
        return resp



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import os

import six
from subprocess import Popen, PIPE

import time

import storops.exception as ex
from storops.lib.common import int_var, text_var, synchronized, cache, daemon

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class NaviCommand(object):
    def __init__(self, username=None, password=None, scope=0,
                 sec_file=None, timeout=None, naviseccli=None):
        self._username = username
        self._password = password
        self._scope = scope
        self._sec_file = sec_file
        self._timeout = timeout
        self._customized_cli = naviseccli
        self._is_credential_valid = True

    MAX_TIMEOUT = 1800
    MIN_TIMEOUT = 3

    def set_credential(self, username=None, password=None, scope=None,
                       sec_file=None):
        if username is not None:
            self._username = username
        if password is not None:
            self._password = password
        if scope is not None:
            self._scope = scope
        if sec_file is not None:
            self._sec_file = sec_file
        self._is_credential_valid = True

    @property
    def timeout(self):
        ret = self._timeout
        if ret is not None:
            ret = int(ret)
            if ret > 1800:
                log.warning('timeout {0} is larger than {1}, reset to {1}.'
                            .format(ret, self.MAX_TIMEOUT))
                ret = 1800
            if ret < 3:
                log.warning('timeout {0} is less than {1}, reset to {1}.'
                            .format(ret, self.MIN_TIMEOUT))
                ret = 3
        return ret

    @property
    def is_credential_valid(self):
        return self._is_credential_valid

    def get_credentials(self):
        if self._sec_file:
            # use security file
            ret = text_var('-secfilepath', self._sec_file)
        elif self._username is None and self._password is None:
            ret = []
        elif self._username is None or self._password is None:
            self._is_credential_valid = False
            raise ex.VNXCredentialError('username or password missing.')
        else:
            ret = ['-user', self._username,
                   '-password', self._password,
                   '-scope', self._scope]
        if self.timeout is not None:
            ret += int_var('-t', self.timeout)
        return ret

    _cli_binary_candidates = (
        r'/opt/Navisphere/bin/naviseccli',
        r'C:\Program Files (x86)\EMC\Navisphere CLI\naviseccli.exe',
        r'C:\Program Files\EMC\Navisphere CLI\naviseccli.exe')

    def set_binary(self, binary):
        """ set the location of naviseccli binary

        :param binary: the absolute path of naviseccli
        :return: nothing
        """
        self._customized_cli = binary

    def _binary(self):
        if self._customized_cli is None:
            binary = 'naviseccli'
            for c in self._cli_binary_candidates:
                if os.path.exists(c):
                    binary = c
                    break
        else:
            binary = self._customized_cli
        self._init_security_level(binary)
        return binary

    def get_cmd_prefix(self, ip):
        binary = self._binary()
        return [binary, '-h', ip] + self.get_credentials()

    @classmethod
    def execute_naviseccli(cls, cmd):
        cmd = list(map(six.text_type, cmd))
        try:
            ret = cls.execute(cmd)
        except OSError:
            raise ex.NaviseccliNotAvailableError()
        return ret

    @classmethod
    def execute(cls, cmd, timeout=None):
        if timeout is None:
            timeout = cls.MAX_TIMEOUT
        # closure cannot modify value, use list to work around
        process = [None]
        output = [None]

        start = time.time()
        cls._log_command(cmd)

        def run():

            p = Popen(cmd, bufsize=-1, stdout=PIPE, stderr=PIPE)
            process[0] = p
            out = p.stdout.read()

            if isinstance(out, bytes):
                out = out.decode("utf-8")
            out = out.strip()
            output[0] = out
            return out

        thread = daemon(run)
        thread.join(timeout)
        cls._log_output(cmd, output[0], start)
        if thread.is_alive() and process[0] is not None:
            log.warn('terminate timeout command: {}'.format(cmd))
            process[0].terminate()
            thread.join()
        return output[0]

    @classmethod
    def _log_command(cls, cmd):
        log.info('call command: {}'.format(cls._get_cmd_str(cmd)))

    @classmethod
    def _get_cmd_str(cls, cmd):
        cmd_cpy = cmd[:]
        # shadow password
        if len(cmd_cpy) >= 7 and cmd_cpy[5] == '-password':
            cmd_cpy[6] = '***'
        cmd_str = ' '.join(cmd_cpy)
        return cmd_str

    @classmethod
    def _log_output(cls, cmd, output, start):
        if log.isEnabledFor(logging.DEBUG):
            output = six.text_type(output).replace('\r\n', '\n').strip()
            if not output:
                output = 'empty'
            else:
                output = '\n' + output

            dt = time.time() - start
            log.debug(
                'command complete: {}, time consumed (s): {}, '
                'output: {}'.format(cls._get_cmd_str(cmd), dt, output))

    @classmethod
    def get_security_level(cls, binary):
        cmd = 'security -certificate -getLevel'.split()
        cmd.insert(0, binary)
        return cls.execute_naviseccli(cmd).lower()

    @classmethod
    def set_security_level(cls, binary, level='low'):
        possible_security_level = ('low', 'high')
        if level not in possible_security_level:
            raise ValueError(
                'possible security level: {}'.format(
                    possible_security_level))
        cmd = 'security -certificate -setLevel {}'.format(level).split()
        cmd.insert(0, binary)
        cls.execute_naviseccli(cmd)

    @staticmethod
    @synchronized()
    @cache
    def _init_security_level(binary):
        # have to specify the specified class
        # otherwise cls could be different for different subclass
        # and cache won't work.
        cls = NaviCommand
        current_level = cls.get_security_level(binary)
        if current_level != 'low':
            log.warn('security level is "{}", update to "low".'.format(
                current_level))
            cls.set_security_level(binary, 'low')



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource

__author__ = 'Yong Huang'

LOG = logging.getLogger(__name__)


class UnityCifsServerInfo(UnityAttributeResource):
    @staticmethod
    def to_embedded(name=None, description=None, netbios_name=None,
                    domain=None, workgroup=None, is_standalone=None,
                    cifs_username=None):
        """
        Constructs an embeded object of `UnityCifsServerInfo`.

        :param name: User-specified name for the SMB server.
        :param description: Description of the SMB server.
        :param netbios_name: Computer Name of the SMB server in Windows
            network.
        :param domain: Domain name where SMB server is registered in
            Active Directory, if applicable.
        :param workgroup: (Applies to stand-alone SMB servers.) Windows
            network workgroup for the SMB server.
        :param is_standalone: Indicates whether the SMB server is standalone.
            Values are:
                true - SMB server is standalone.
                false - SMB server is joined to the Active Directory.
        :param cifs_username: User name for authentication to CIFS server
            on the source VDM. it could be the same as Active Directory
            domain user name in some cases, and which will be used
            initialized to the Active Directory domain user name in GUI side.
        :return:
        """
        return {'name': name, 'description': description,
                'netbiosName': netbios_name, 'domain': domain,
                'workgroup': workgroup, 'isStandalone': is_standalone,
                'cifsUsername': cifs_username}


class UnityImportSyncProgress(UnityResource):
    pass


class UnityImportSyncProgressList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityImportSyncProgress


class UnityElementImport(UnityResource):
    pass


class UnityElementImportList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityElementImport


class UnityImportSession(UnityResource):
    @classmethod
    def create_block_import(cls, cli, remote_system, src_resource,
                            dst_resource_config,
                            dst_resource_element_configs=None,
                            import_as_vmware_datastore=None, throttle=None,
                            hosts=None, cutover_threshold=None, name=None):
        """
        Create a block import session.

        :param remote_system: Source system from which the import is being
            configured.
        :param src_resource: Unique identifier of the source resource.
        :param dst_resource_config: This is the user chosen config for
            destination resource provisioning.
        :param dst_resource_element_configs: This is the user chosen config
            for each of the member element of the destination resource.
        :param import_as_vmware_datastore: Specify if the target resource
            should be VMware Datastore. Default is 'false' and not
            application to consistency groups, only standalone luns.
        :param throttle: Throttle for session. 'True' (Default):
            Session throttled. 'False': Session not throttled.
        :param hosts: List of hosts that will be granted access to
            import session destination resources.
        :param cutover_threshold: Cutover threshold. Default value: 5
        :param name: Name of the import session.
        :return: The newly created import session.
        """

        req_body = cli.make_body(
            remoteSystem=remote_system,
            srcResource=src_resource,
            dstResourceConfig=dst_resource_config,
            dstResourceElementConfigs=dst_resource_element_configs,
            importAsVMwareDatastore=import_as_vmware_datastore,
            throttle=throttle, hosts=hosts,
            cutoverThreshold=cutover_threshold, name=name)

        resp = cli.type_action(
            cls().resource_class,
            'createBlockImport',
            **req_body)
        resp.raise_if_err()
        # response is like:
        # "content": {
        #     "id": {
        #         "id": "import_xxxx"
        #     }
        session_resp = resp.first_content['id']
        return cls.get(cli, _id=session_resp['id'])

    def modify(self, throttle=None, name=None, cutover_threshold=None,
               src_local_cifs_admin_username=None,
               src_local_cifs_admin_passwd=None):
        """
        Modify an import session.

        :param throttle: Throttle for session. 'True': Session throttled.
            'False': Session not throttled.
        :param name: Name of the import session.
        :param cutover_threshold: Cutover threshold.
        :param src_local_cifs_admin_username: User name for authentication
            to CIFS server on the source VDM. It is only used for VDM import.
        :param src_local_cifs_admin_passwd: Password for authentication
            to CIFS server on the source VDM. It is only used for VDM import.
        :return:
        """
        req_body = self._cli.make_body(
            throttle=throttle, name=name, cutoverThreshold=cutover_threshold,
            srcLocalCifsAdminUsername=src_local_cifs_admin_username,
            srcLocalCifsAdminPasswd=src_local_cifs_admin_passwd)
        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp

    def pause(self, test_facility_code=None):
        """
        Pause an import session.
        """
        req_body = self._cli.make_body(testFacilityCode=test_facility_code)
        resp = self.action('pause', **req_body)
        resp.raise_if_err()
        return resp

    def resume(self):
        """
        Resume an import session.
        """
        resp = self.action('resume')
        resp.raise_if_err()
        return resp

    def cutover(self, cifs_server_info=None, domain_username=None,
                domain_password=None):
        """
        Cutover an import session.

        For block import this switches the production over as well as commit
        the session. For file import this just switches the production over
        to destination site.

        :param cifs_server_info: New CIFS server object to be created at
            the source VDM.
        :param domain_username: Active Directory domain user name.
        :param domain_password: Active Directory domain password.
        :return:
        """
        req_body = self._cli.make_body(cifsServerInfo=cifs_server_info,
                                       domainUsername=domain_username,
                                       domainPassword=domain_password)
        resp = self.action('cutover', **req_body)
        resp.raise_if_err()
        return resp

    def commit(self):
        """
        Commit an import session.
        """
        resp = self.action('commit')
        resp.raise_if_err()
        return resp

    def cancel(self, domain_username=None, domain_password=None):
        """
        Cancel an import session.

        :param domain_username: Active Directory domain user name.
        :param domain_password: Active Directory domain password.
        :return:
        """
        req_body = self._cli.make_body(domainUsername=domain_username,
                                       domainPassword=domain_password)
        resp = self.action('cancel', **req_body)
        resp.raise_if_err()
        return resp


class UnityImportSessionList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityImportSession



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import functools
import logging
import os

import six
import xmltodict as xmltodict
import xml.etree.ElementTree as ET
from mock import patch

from storops.lib.common import cache, allow_omit_parentheses
from storops.vnx.nas_client import VNXNasClient
from storops.vnx.xmlapi import XML_NS
from storops_test.utils import ConnectorMock, read_test_file
from storops_test.vnx.cli_mock import MockCli

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


ET.register_namespace('', XML_NS)


@cache
def t_nas():
    """ get the test NAS client

    :return: test NAS client
    """
    return VNXNasClient("10.110.24.191", 'nasadmin', 'nasadmin')


class MockXmlPost(ConnectorMock):
    base_folder = os.path.join('vnx', 'testdata', 'nas_xml_output')

    @classmethod
    def get_folder(cls, body):
        skipped_nodes = ('RequestPacket', 'Request')

        ret = [cls.base_folder]
        # get two levels after Request
        node = ET.fromstring(body.encode('utf-8'))
        while len(ret) < 3:
            tag = cls.delete_ns(node.tag)
            if tag in skipped_nodes:
                pass
            else:
                ret.append(tag)
            children = node.getchildren()
            if not children:
                break
            node = children[0]
        return os.path.join(*ret)

    def mock_post(self, body):
        return self.get_mock_output(body)

    @staticmethod
    def delete_ns(tag):
        if '}' in tag:
            tag = tag[tag.index('}') + 1:]
        return tag

    @staticmethod
    @cache
    def read_index(folder):
        return read_test_file(folder, 'index.xml')

    @classmethod
    def get_filename(cls, body):
        xml_string = cls.read_index(cls.get_folder(body))
        indices = ET.fromstring(xml_string.encode('utf-8'))
        ret = None
        for index in indices:
            for request_packet in index:
                if xml_compare(request_packet, body):
                    if 'output' not in index.attrib:
                        raise AttributeError(
                            'missing "output" attribute in "Index" node.')
                    ret = index.attrib['output']
                    break
            if ret:
                break
        else:
            log.error('cannot find response for request: \n{}'.format(body))

        return ret


@allow_omit_parentheses
def patch_post(output=None, mock_map=None):
    xml = MockXmlPost(output, mock_map)

    def decorator(func):
        @functools.wraps(func)
        @patch(target='storops.connection.connector.XMLAPIConnector.post',
               new=xml.mock_post)
        @patch(target='storops.connection.connector.XMLAPIConnector._login',
               new=lambda a, b, c, d: 0)
        @patch(target='storops.lib.common.const_seconds',
               new=lambda x: 0)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

    return decorator


class MockSsh(MockCli):
    base_folder = os.path.join('vnx', 'testdata', 'ssh_output')

    flags_to_delete = {'NAS_DB=/nas': 1}

    def mock_execute(self, command, timeout=None, check_exit_code=True):
        if len(command) > 2:
            folder = '/nas/bin/'
            if folder in command[2]:
                command[2] = command[2].replace(folder, '')
        return self.get_mock_output(command), None

    def mock_init_connection(self, *args):
        pass


def patch_ssh(output=None, mock_map=None):
    ssh = MockSsh(output, mock_map)

    def decorator(func):
        @functools.wraps(func)
        @patch(target='storops.connection.connector.SSHConnector.execute',
               new=ssh.mock_execute)
        @patch(target='storops.connection.connector.SSHConnector.'
                      'init_connection',
               new=ssh.mock_init_connection)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

    return decorator


@allow_omit_parentheses
def patch_nas(xml_output=None, xml_map=None, ssh_output=None, ssh_map=None):
    xml = MockXmlPost(xml_output, xml_map)
    ssh = MockSsh(ssh_output, ssh_map)

    def decorator(func):
        @functools.wraps(func)
        @patch(target='storops.connection.connector.XMLAPIConnector.post',
               new=xml.mock_post)
        @patch(target='storops.connection.connector.XMLAPIConnector._login',
               new=lambda a, b, c, d: 0)
        @patch(target='storops.lib.common.const_seconds',
               new=lambda x: 0)
        @patch(target='storops.connection.connector.SSHConnector.execute',
               new=ssh.mock_execute)
        @patch(target='storops.connection.connector.SSHConnector.'
                      'init_connection',
               new=ssh.mock_init_connection)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

    return decorator


def normalise_dict(d):
    """ Recursively convert dict-like object (eg OrderedDict) into plain dict.

    And sorts list values.
    :param d: dict input
    """
    out = {}
    for k, v in d.items():
        if hasattr(v, 'items'):
            out[k] = normalise_dict(v)
        elif isinstance(v, list):
            out[k] = []
            for item in sorted(v):
                if hasattr(item, 'items'):
                    out[k].append(normalise_dict(item))
                else:
                    out[k].append(item)
        else:
            out[k] = v
    return out


def xml_compare(a, b):
    """ Compares two XML documents (as string or etree)

    Does not care about element order
    :param a: xml to compare
    :param b: xml to compare
    """
    if not isinstance(a, six.string_types):
        a = ET.tostring(a, encoding='utf-8').decode('utf-8')
    if not isinstance(b, six.string_types):
        b = ET.tostring(b, encoding='utf-8').decode('utf-8')
    a = normalise_dict(xmltodict.parse(a))
    b = normalise_dict(xmltodict.parse(b))
    return a == b



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import pytest

from storops_comptest.utils import setup_fixture
from storops_comptest.vnx import VNXGeneralFixtureManager, \
    MultiVNXGeneralFixtureManager

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


@pytest.fixture(scope='session')
def vnx_gf(request):
    """ General fixture for most vnx cases

        Details including:
            vnx  - reference to the system.
            pool - A RAID5 pool with 3 disks created on the fly.
            lun  - A LUN created in the pool.
            snap - A snap created upon the LUN.
    :param request:
    :return:
    """
    return setup_fixture(request, VNXGeneralFixtureManager)


@pytest.fixture(scope='session')
def multi_vnx_gf(request):
    """ general fixture for multi VNX test cases

        Details including:
            vnx - reference to the system
            sync_mirror - a synchronized mirror

    :param request:
    :return:
    """
    return setup_fixture(request, MultiVNXGeneralFixtureManager)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of

from storops.unity.enums import StorageResourceTypeEnum, ReplicationTypeEnum, \
    ThinStatusEnum, TieringPolicyEnum
from storops.unity.resource.filesystem import UnityFileSystem
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.pool import UnityPoolList
from storops.unity.resource.storage_resource import UnityStorageResource, \
    UnityStorageResourceList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


class UnityStorageResourceTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        sr = UnityStorageResource(_id='res_27', cli=t_rest())
        assert_that(sr.id, equal_to('res_27'))
        assert_that(sr.type, equal_to(StorageResourceTypeEnum.FILE_SYSTEM))
        assert_that(sr.replication_type, equal_to(ReplicationTypeEnum.NONE))
        assert_that(sr.thin_status, equal_to(ThinStatusEnum.TRUE))
        assert_that(sr.relocation_policy,
                    equal_to(TieringPolicyEnum.AUTOTIER_HIGH))
        assert_that(sr.health, instance_of(UnityHealth))
        assert_that(sr.name, equal_to('fs3'))
        assert_that(sr.description, equal_to(''))
        assert_that(sr.is_replication_destination, equal_to(False))
        assert_that(sr.size_total, equal_to(3221225472))
        assert_that(sr.size_used, equal_to(1620303872))
        assert_that(sr.size_allocated, equal_to(3221225472))
        assert_that(sr.per_tier_size_used, equal_to([6442450944, 0, 0]))
        assert_that(sr.metadata_size, equal_to(3489660928))
        assert_that(sr.metadata_size_allocated, equal_to(3221225472))
        assert_that(sr.snaps_size_total, equal_to(0))
        assert_that(sr.snaps_size_allocated, equal_to(0))
        assert_that(sr.snap_count, equal_to(0))
        assert_that(sr.pools, instance_of(UnityPoolList))
        assert_that(sr.filesystem, instance_of(UnityFileSystem))

    @patch_rest
    def test_get_all(self):
        sr_list = UnityStorageResourceList(cli=t_rest())
        assert_that(len(sr_list), equal_to(10))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from hamcrest import assert_that, equal_to, has_items, has_item, none

from storops_comptest.utils import is_jenkins
from storops import HostTypeEnum, ACEAccessLevelEnum

__author__ = 'Cedric Zhuang'


def test_check_cifs_share_exists(unity_gf):
    assert_that(unity_gf.cifs_share.existed, equal_to(True))


def test_create_delete_cifs_share_snapshot(unity_gf):
    # create snap
    snap = unity_gf.cifs_share.create_snap(unity_gf.add_snap_name())
    assert_that(unity_gf.cifs_share.storage_resource.id,
                equal_to(snap.storage_resource.id))

    # delete snap
    resp = snap.delete()
    snap.update()
    assert_that(resp.is_ok(), equal_to(True))
    assert_that(snap.existed, equal_to(False))


def test_check_nfs_share_exists(unity_gf):
    assert_that(unity_gf.nfs_share.existed, equal_to(True))


def test_create_delete_nfs_share_snapshot(unity_gf):
    # create snap
    snap = unity_gf.nfs_share.create_snap(unity_gf.add_snap_name())
    assert_that(unity_gf.nfs_share.storage_resource.id,
                equal_to(snap.storage_resource.id))

    # delete snap
    resp = snap.delete()
    snap.update()
    assert_that(resp.is_ok(), equal_to(True))
    assert_that(snap.existed, equal_to(False))


def test_create_snap_based_nfs_share_and_snap(unity_gf):
    snap = unity_gf.nfs_share.create_snap(unity_gf.add_snap_name())

    # create snap based share
    nfs_share = snap.create_nfs_share(name=snap.name)
    assert_that(nfs_share.existed, equal_to(True))

    # create snap of snap based share
    share_snap = nfs_share.create_snap(unity_gf.add_snap_name())
    assert_that(share_snap.existed, equal_to(True))


def test_create_snap_based_cifs_share_and_snap(unity_gf):
    snap = unity_gf.cifs_share.create_snap(unity_gf.add_snap_name())

    # create snap based share
    cifs_share = snap.create_cifs_share(name=snap.name)
    assert_that(cifs_share.existed, equal_to(True))

    # create snap of snap based share
    share_snap = cifs_share.create_snap(unity_gf.add_snap_name())
    assert_that(share_snap.existed, equal_to(True))


def test_nfs_share_read_write_host_access(unity_gf):
    share = unity_gf.create_nfs_share()
    share.allow_read_write_access(['1.1.1.1', '1.1.1.2'],
                                  force_create_host=True)

    share.update()
    assert_that(share.read_write_hosts.ip_list,
                has_items('1.1.1.1', '1.1.1.2'))

    share.deny_access(['1.1.1.1', '1.1.1.2'])
    share.update()
    assert_that(share.no_access_hosts.ip_list, has_items('1.1.1.1', '1.1.1.2'))


def test_nfs_share_mixed_host_access(unity_gf):
    share = unity_gf.nfs_share
    share.allow_read_only_access('1.1.1.3', force_create_host=True)
    share.allow_root_access('1.1.1.4', force_create_host=True)
    share.update()
    assert_that(share.read_only_hosts.ip_list, has_item('1.1.1.3'))
    assert_that(share.root_access_hosts.ip_list, has_item('1.1.1.4'))

    share.deny_access(['1.1.1.3', '1.1.1.4'])
    share.update()
    assert_that(share.read_only_hosts, none())
    assert_that(share.root_access_hosts, none())
    assert_that(share.no_access_hosts.ip_list, has_items('1.1.1.3', '1.1.1.4'))


def test_deny_access_to_subnet(unity_gf):
    share = unity_gf.nfs_share
    share.deny_access('7.7.7.7/8', force_create_host=True)

    share.update()
    assert_that(share.no_access_hosts.ip_list, has_items('7.7.7.7'))

    host = unity_gf.unity.get_host(address='7.7.7.7')
    assert_that(host.type, equal_to(HostTypeEnum.SUBNET))


@pytest.mark.skipif(is_jenkins(), reason='No run on CI, manual only.')
def test_cifs_access_control(unity_cs):
    share = unity_cs.cifs_share
    resp = share.enable_ace()
    assert_that(resp.is_ok(), equal_to(True))

    access_list = share.get_ace_list()
    orig = len(access_list[ACEAccessLevelEnum.FULL])

    share.add_ace('win2012.dev', 'SMIS_User_1')
    access_list = share.get_ace_list()
    assert_that(len(access_list[ACEAccessLevelEnum.FULL]), equal_to(orig + 1))

    share.delete_ace('win2012.dev', 'SMIS_User_1')
    access_list = share.get_ace_list()
    assert_that(len(access_list[ACEAccessLevelEnum.FULL]), equal_to(orig))

    resp = share.disable_ace()
    assert_that(resp.is_ok(), equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import storops.vnx.resource.nfs_share

from storops.lib.common import check_int
from storops.vnx.enums import VNXPortType
from storops.vnx.resource import VNXCliResourceList, VNXResource

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class VNXMoverRefList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverRef

    def _get_raw_resource(self):
        return self._cli.get_mover(full=False)


class VNXMoverRef(VNXResource):
    _full_prop = False

    def __init__(self, name=None, mover_id=None, cli=None):
        super(VNXMoverRef, self).__init__()
        self._name = name
        self._mover_id = mover_id
        self._cli = cli

        # cache for members
        self._host = None

    def update(self, data=None):
        super(VNXMoverRef, self).update(data)

    @property
    def is_vdm(self):
        return False

    def get_mover_id(self):
        if self._mover_id is not None:
            ret = self._mover_id
        else:
            ret = self.mover_id
        return ret

    def _get_raw_resource(self):
        if self._mover_id is not None:
            resp = self._cli.get_mover(mover_id=self._mover_id,
                                       full=self._full_prop)
        elif self._name is not None:
            resp = self._cli.get_mover(full=self._full_prop)
            resp.filter_object(name=self._name)
        else:
            raise ValueError('mover_id or name should be supplied.')
        return resp

    @classmethod
    def get_id(cls, item):
        if isinstance(item, VNXMoverRef):
            ret = item.get_mover_id()
        else:
            ret = check_int(item)
        return ret

    def create_dns(self, domain_name, ip):
        if isinstance(ip, (list, tuple)):
            ip = ' '.join(ip)
        resp = self._cli.create_dns_domain(self.get_mover_id(), domain_name,
                                           ip)
        resp.raise_if_err()
        return resp

    def delete_dns(self, domain_name):
        resp = self._cli.delete_dns_domain(self.get_mover_id(), domain_name)
        resp.raise_if_err()
        return resp

    @property
    def host(self):
        if self._host is None:
            self._host = VNXMoverHost(host_id=self.host_id, cli=self._cli)
        return self._host

    @property
    def physical_devices(self):
        return self.host.physical_device

    @property
    def fc_devices(self):
        return [device for device in self.physical_devices
                if device.type == VNXPortType.FC]

    @property
    def ethernet_devices(self):
        return [device for device in self.physical_devices
                if device.type == VNXPortType.ETHERNET]

    def get_interconnect_id(self, source=None, destination=None):
        if source is None:
            source = self.name
        if destination is None:
            destination = self.name

        out = self._cli.get_mover_interconnect_id_list()

        for line in out.strip().split('\n'):
            _id, name, src, dest_system, dest = line.strip().split()
            if src == source and dest == destination:
                conn_id = check_int(_id)
                break
        else:
            conn_id = None

        return conn_id

    def create_nfs_share(self, path, ro=False, host_config=None):
        share_clz = storops.vnx.resource.nfs_share.VNXNfsShare
        return share_clz.create(self._cli, self, path, ro, host_config)


class VNXMoverList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMover

    def _get_raw_resource(self):
        return self._cli.get_mover()


class VNXMover(VNXMoverRef):
    _full_prop = True

    @staticmethod
    def get(cli, name=None, mover_id=None):
        if name is not None or mover_id is not None:
            ret = VNXMover(name=name, mover_id=mover_id, cli=cli)
        else:
            ret = VNXMoverList(cli=cli)
        return ret

    @property
    def interfaces(self):
        ret = []
        for i in self.mover_interfaces:
            i._mover = self
            i._cli = self._cli
            ret.append(i)
        return ret

    def create_interface(self, device, ip, net_mask, vlan_id=0, name=None):
        mover_id = self.get_mover_id()
        if hasattr(device, 'name'):
            device = device.name
        resp = self._cli.create_mover_interface(
            mover_id, device, ip, net_mask, vlan_id, name)
        resp.raise_if_err()
        return VNXMoverInterface(mover=self, cli=self._cli, ip=ip)

    def delete_interface(self, ip):
        interface = next(i for i in self.interfaces if i.ip_addr == ip)
        return interface.delete()


class VNXMoverInterfaceList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverInterface


class VNXMoverInterface(VNXResource):
    def __init__(self, mover=None, ip=None, cli=None):
        super(VNXMoverInterface, self).__init__()
        self._cli = cli
        self._mover = mover
        self._ip = ip

    def _get_raw_resource(self):
        self._mover.update()
        interface = next(i for i in self._mover.interfaces
                         if i.ip_addr == self._ip)
        return interface.parsed_resource

    def get_ip(self):
        if self._ip is not None:
            ret = self._ip
        else:
            ret = self.ip_addr
        return ret

    def delete(self):
        mover_id = self._mover.get_mover_id()
        resp = self._cli.delete_mover_interface(mover_id, self.get_ip())
        resp.raise_if_err()
        return resp


class VNXMoverLogicalNetworkDeviceList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverLogicalNetworkDevice


class VNXMoverLogicalNetworkDevice(VNXResource):
    pass


class VNXMoverRouteList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverRoute


class VNXMoverRoute(VNXResource):
    pass


class VNXMoverDeduplicationSettings(VNXResource):
    pass


class VNXMoverHostList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverHost

    def _get_raw_resource(self):
        return self._cli.get_mover_host()


class VNXMoverHost(VNXResource):
    def __init__(self, host_id=None, cli=None):
        super(VNXMoverHost, self).__init__()
        self._host_id = host_id
        self._cli = cli

    def _get_raw_resource(self):
        return self._cli.get_mover_host(self._host_id)


class VNXMoverMotherboard(VNXResource):
    pass


class VNXMoverPhysicalDeviceList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMoverPhysicalDevice


class VNXMoverPhysicalDevice(VNXResource):
    @property
    def is_internal(self):
        ret = False
        if self.name:
            header = self.name[:3]
            ret = header in ('mge', 'fxg', 'tks', 'fsn')
        return ret



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.lib.common import instance_cache, clear_instance_cache
from storops.unity.calculator import IdValues
from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Cedric Zhuang'


class UnityMetric(UnityResource):
    pass


class UnityMetricList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMetric


class UnityMetricRealTimeQuery(UnityResource):
    @classmethod
    def get_query_list(cls, cli, interval, paths):
        queries = UnityMetricRealTimeQueryList(cli=cli, interval=interval)
        queries.sort_by_path()
        paths = set(paths)
        id_list = []
        for query in queries:
            query_paths = set(query.paths)
            if query_paths.intersection(paths):
                id_list.append(query.get_id())
                paths -= query_paths
        if paths:
            id_list.append(
                cls.create(cli, interval, sorted(list(paths))).get_id())
            queries.update()
        return queries.set_id_list(id_list)

    @classmethod
    def create(cls, cli, interval, paths):
        resp = cli.post(cls().resource_class, interval=interval, paths=paths)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def get_query_result(self, paths=None):
        results = UnityMetricQueryResultList(
            cli=self._cli, query_id=self.get_id())
        return results.filtered_by_path(paths)


class UnityMetricRealTimeQueryList(UnityResourceList):
    def __init__(self, cli=None, interval=None, id_list=None):
        super(UnityMetricRealTimeQueryList, self).__init__(cli)
        self._interval = interval
        self._id_list = id_list

    def set_id_list(self, id_list):
        self._id_list = id_list
        self._apply_filter()
        return self

    @classmethod
    def get_resource_class(cls):
        return UnityMetricRealTimeQuery

    def sort_by_path(self, reverse=True):
        if self._list is None:
            self.update()
        if self._list is not None:
            self._list.sort(key=lambda q: len(q.paths), reverse=reverse)

    def _filter(self, item):
        ret = True
        if self._interval is not None:
            ret &= item.interval == self._interval
        if self._id_list is not None:
            ret &= item.id in self._id_list

        return ret

    def get_query_result(self, paths=None):
        ret = None
        for query in self:
            result = query.get_query_result(paths)
            if ret is None:
                ret = result
            else:
                ret.merge(result)
        return ret


class UnityMetricQueryResult(UnityResource):
    @instance_cache
    def sum_sp(self):
        return sum(self.numeric_values)

    @property
    @instance_cache
    def numeric_values(self):
        if self.values is None:
            ret = []
        else:
            ret = [IdValues({k: int(v) for k, v in value.items()})
                   for value in self.values.values()]
        return ret

    @property
    def sp_values(self):
        if self.values is None:
            ret = IdValues()
        else:
            ret = IdValues({k: int(v) for k, v in self.values.items()})
        return ret

    @property
    def sp_sum_values(self):
        """
        return sp level values

        input:
        "values": {
            "spa": {
                "19": "385",
                "18": "0",
                "20": "0",
                "17": "0",
                "16": "0"
            },
            "spb": {
                "19": "101",
                "18": "101",
                "20": "101",
                "17": "101",
                "16": "101"
            }
        },

        return:
        "values": {
            "spa": 385,
            "spb": 505
        },
        """
        if self.values is None:
            ret = IdValues()
        else:
            ret = IdValues({k: sum(int(x) for x in v.values()) for k, v in
                            self.values.items()})
        return ret

    @property
    def sum_sp_values(self):
        """
        return system level values (spa + spb)

        input:
        "values": {
            "spa": 385,
            "spb": 505
        },

        return:
        "values": {
            "0": 890
        },
        """
        if self.values is None:
            ret = IdValues()
        else:
            ret = IdValues({'0': sum(int(x) for x in self.values.values())})
        return ret

    def combine_numeric_values(self, other):
        """
        numeric_values * sp_values
        """
        if self.values is None:
            ret = IdValues()
        else:
            ret = sum([IdValues(
                {k: int(v) * int(other.values[key]) for k, v in value.items()})
                for key, value in self.values.items()])
        return ret

    def combine_sp_values(self, other):
        """
        sp_values * sp_values
        """
        if self.values is None:
            ret = IdValues()
        else:
            ret = IdValues({k: int(v) * int(other.values[k]) for k, v in
                            self.values.items()})
        return ret

    def sum_combined_sp_values(self, other):
        """
        sum(sp_values * sp_values)
        """
        if self.values is None:
            ret = IdValues()
        else:
            ret = IdValues({'0': sum(int(x) for x in
                                     {k: int(v) * int(other.values[k]) for k, v
                                      in self.values.items()}.values())})
        return ret

    def diff_timestamp(self, other):
        if other is None:
            ret = None
        else:
            ret = self.timestamp - other.timestamp
        return ret

    def diff_seconds(self, other):
        if other is None:
            ret = None
        else:
            ret = abs(self.diff_timestamp(other).total_seconds())
        return ret


class UnityMetricQueryResultList(UnityResourceList):
    def __init__(self, cli=None, **the_filter):
        super(UnityMetricQueryResultList, self).__init__(cli, **the_filter)
        self._path_result_map = {}

    @classmethod
    def get_resource_class(cls):
        return UnityMetricQueryResult

    @clear_instance_cache
    def update(self, data=None):
        super(UnityMetricQueryResultList, self).update(data)
        self._path_result_map = {}

    def by_path(self, path):
        if not self._path_result_map:
            for r in self:
                self._path_result_map[r.path] = r
        return self._path_result_map.get(path)

    def diff_seconds(self, other):
        if other is None or len(other) == 0 or len(self) == 0:
            ret = None
        else:
            ret = self[0].diff_seconds(other[0])
        return ret

    def filtered_by_path(self, paths=None):
        if paths is not None:
            self._list = [q for q in self if q.path in paths]
        return self

    def merge(self, other):
        if other is not None:
            my_paths = self.path
            to_add = [r for r in other if r.path not in my_paths]
            if self._list is None:
                self._list = []
            self._list.extend(to_add)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from storops.lib.common import check_int
import storops.vnx.resource.fs
from storops.vnx.resource import VNXResource, VNXCliResourceList

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXNasPoolList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXNasPool

    def _get_raw_resource(self):
        return self._cli.get_nas_pool()


class VNXNasPool(VNXResource):
    def __init__(self, name=None, pool_id=None, cli=None):
        super(VNXNasPool, self).__init__()
        self._name = name
        self._pool_id = pool_id
        self._cli = cli

    @classmethod
    def get(cls, cli, name=None, pool_id=None):
        if name is not None or pool_id is not None:
            ret = VNXNasPool(name=name, pool_id=pool_id, cli=cli)
        else:
            ret = VNXNasPoolList(cli=cli)
        return ret

    def get_pool_id(self):
        if self._pool_id is not None:
            ret = self._pool_id
        else:
            ret = self.pool_id
        return ret

    @staticmethod
    def get_id(pool):
        if isinstance(pool, VNXNasPool):
            pool = pool.get_pool_id()
        try:
            ret = check_int(pool)
        except ValueError:
            raise ValueError('invalid pool id supplied: {}'
                             .format(pool))
        return ret

    def _get_raw_resource(self):
        resp = self._cli.get_nas_pool()
        resp.filter_object(name=self._name, pool=self._pool_id)
        return resp

    def create_filesystem(self, name, size, mover=1, is_vdm=False):
        fs = storops.vnx.resource.fs
        fs.VNXFileSystem.create(self._cli, name, self, size, mover, is_vdm)
        return fs.VNXFileSystem(name=name, cli=self._cli)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, equal_to, only_contains, none, any_of, \
    contains_string, raises

from storops.unity.client import UnityClient, UnityDoc
from storops.unity.enums import RaidTypeEnum, HealthEnum, RaidTypeEnumList, \
    ServiceLevelEnum, ServiceLevelEnumList
from storops.unity.resource.lun import UnityLun, UnityLunList
from storops_test.unity.rest_mock import patch_rest, t_rest

__author__ = 'Cedric Zhuang'


class UnityClientTest(unittest.TestCase):
    def test_assemble_url_no_compact(self):
        url = UnityClient.assemble_url('api/types', filter='a eq 100')
        assert_that(url, equal_to('/api/types?compact=True&filter=a eq 100'))

    def test_assemble_url_none_filter(self):
        url = UnityClient.assemble_url('api/types', filter=None)
        assert_that(url, equal_to('/api/types?compact=True'))

    @patch_rest
    def test_get_fields(self):
        fields = t_rest().get_fields('metric')
        assert_that(fields, only_contains(
            'id', 'name', 'path', 'product', 'type', 'objectType',
            'description', 'isHistoricalAvailable', 'isRealtimeAvailable',
            'unitDisplayString', 'unit', 'metricGroupName', 'visibility'))

    @patch_rest(output='abc.json')
    def test_mock_file_not_found(self):
        def f():
            t_rest().get_fields('abc')

        assert_that(f, raises(IOError))

    @patch_rest
    def test_make_body_complex(self):
        service_levels = [ServiceLevelEnum.BASIC, ServiceLevelEnum.BRONZE]
        param = {
            'a': 1,
            'b': UnityLun(_id='lun1'),
            'c': UnityLunList(cli=t_rest()),
            'd': [UnityLun(_id='lun10'), UnityLun(_id='lun11'), 0.1],
            'e': {'f': UnityLun(_id='lun12')},
            'g': 'string',
            'h': 0.2,
            'i': service_levels,
            'j': ServiceLevelEnumList.parse(service_levels)
        }
        ret = UnityClient.make_body(param)
        expected = {'a': 1,
                    'b': {'id': 'lun1'},
                    'c': [{'id': 'sv_2'}, {'id': 'sv_3'},
                          {'id': 'sv_5'}, {'id': 'sv_6'},
                          {'id': 'sv_7'}],
                    'd': [{'id': 'lun10'}, {'id': 'lun11'}, 0.1],
                    'e': {'f': {'id': 'lun12'}},
                    'g': 'string', 'h': 0.2, 'i': [0, 1], 'j': [0, 1]}
        assert_that(ret, equal_to(expected))

    @patch_rest
    def test_make_body_blank(self):
        param = {'fastVPParameters': {
            'tieringPolicy': None
        }}
        ret = UnityClient.make_body(param)
        expected = {}
        assert_that(ret, equal_to(expected))

    def test_make_body_nested_empty_dict(self):
        param = {
            'name': 'abc',
            'replicationParameters': {
                'isReplicationDestination': None,
            }
        }
        ret = UnityClient.make_body(param)
        assert_that(ret, equal_to({'name': 'abc'}))

    def test_make_body_no_change(self):
        ret = UnityClient.make_body(True)
        assert_that(ret, equal_to(True))

        ret = UnityClient.make_body('string')
        assert_that(ret, equal_to('string'))

    def test_make_body_resource(self):
        ret = UnityClient.make_body(UnityLun(_id='abc'))
        assert_that(ret, equal_to({'id': 'abc'}))

    def test_make_body_None(self):
        ret = UnityClient.make_body({'a': None})
        assert_that(ret, equal_to({}))

    def test_make_body_enum(self):
        ret = UnityClient.make_body({'a': RaidTypeEnum.RAID5})
        assert_that(ret, equal_to({'a': 1}))

    def test_make_body_kwargs(self):
        ret = UnityClient.make_body(a=1, b='c')
        assert_that(ret, equal_to({'a': 1, 'b': 'c'}))

    def test_make_body_zero(self):
        ret = UnityClient.make_body(a=0, b='')
        assert_that(ret, equal_to({'a': 0, 'b': ''}))

    def test_make_body_empty_collection(self):
        ret = UnityClient.make_body(allow_empty=True, b=(), c=[])
        assert_that(ret, equal_to({'b': [], 'c': []}))

    def test_make_body_nested_empty_collection(self):
        ret = UnityClient.make_body(allow_empty=True, b=(), c={'d': []})
        assert_that(ret, equal_to({'b': [], 'c': {'d': []}}))

    def test_make_body_empty_dict(self):
        inner = UnityClient.make_body(a=None)
        outer = UnityClient.make_body(b=inner, c=3)
        assert_that(outer, equal_to({'c': 3}))

    def test_dict_to_filter_string_normal(self):
        ret = UnityClient.dict_to_filter_string({'a': 1, 'b': 'c'})
        assert_that(ret, equal_to('a eq 1 and b eq "c"'))

    def test_dict_to_filter_unity_resource(self):
        ret = UnityClient.dict_to_filter_string(
            {'c': 1, 'b': UnityLun(_id='lun_1')})
        assert_that(ret, equal_to('b eq "lun_1" and c eq 1'))

    def test_dict_to_filter_list(self):
        ret = UnityClient.dict_to_filter_string({'a': [2, 4]})
        assert_that(ret, any_of('a eq 2 or a eq 4', 'a eq 4 or a eq 2'))
        ret = UnityClient.dict_to_filter_string({'a': ["m", "n"]})
        assert_that(ret, any_of('a eq "m" or a eq "n"',
                                'a eq "n" or a eq "m"'))

    def test_dict_to_filter_string_enum_list(self):
        levels = [ServiceLevelEnum.SILVER, ServiceLevelEnum.PLATINUM]
        ret = UnityClient.dict_to_filter_string({'a': levels})
        assert_that(ret, any_of('a eq 2 or a eq 4', 'a eq 4 or a eq 2'))

    def test_dict_to_filter_string_value_none(self):
        ret = UnityClient.dict_to_filter_string({'a': None, 'b': 'c'})
        assert_that(ret, equal_to('b eq "c"'))

    def test_dict_to_filter_string_value_all_none(self):
        ret = UnityClient.dict_to_filter_string({'a': None, 'b': None})
        assert_that(ret, none())

    def test_dict_to_filter_string_empty(self):
        ret = UnityClient.dict_to_filter_string({})
        assert_that(ret, none())
        ret = UnityClient.dict_to_filter_string(None)
        assert_that(ret, none())

    @patch_rest
    def test_system_version(self):
        assert_that(t_rest().system_version, equal_to('4.1.0'))


class UnityDocTest(unittest.TestCase):
    @patch_rest
    def test_get_doc_of_field(self):
        unity_doc = UnityDoc(t_rest(), UnityLun)
        doc = unity_doc._get_doc(field='name')
        assert_that(doc, equal_to('Readable name'))

    @patch_rest
    def test_get_doc_of_resource(self):
        unity_doc = UnityDoc(t_rest(), UnityLun)
        doc = unity_doc._get_doc()
        assert_that(doc, contains_string('Represents Volume'))

    @patch_rest
    def test_get_doc_of_index(self):
        unity_doc = UnityDoc(t_rest(), HealthEnum)
        doc = unity_doc._get_doc(value=5)
        assert_that(doc, equal_to('OK'))

    @patch_rest
    def test_get_doc_enum_list(self):
        unity_doc = UnityDoc(t_rest(), RaidTypeEnumList)
        doc = unity_doc.doc
        assert_that(doc, contains_string('RAID5 has the only'))

    @patch_rest
    def test_get_doc_unity_resource_list(self):
        unity_doc = UnityDoc(t_rest(), UnityLunList)
        doc = unity_doc.doc
        assert_that(doc, contains_string('Represents Volume'))
        assert_that(doc, contains_string('current_node'))
        assert_that(doc, contains_string('Current SP'))

    def test_get_fmt_str(self):
        assert_that(UnityDoc.get_fmt_str((3, 5, 2)), equal_to('{:5}{:7}{:4}'))

    def test_get_column_max_len_str(self):
        data = [('but', '', 'd'), ('of', 'hand', 'i')]
        assert_that(UnityDoc.get_column_max_len(data), equal_to([3, 4, 1]))

    def test_get_column_max_len_int(self):
        data = [(1, 'd'), (11, 'd')]
        assert_that(UnityDoc.get_column_max_len(data), equal_to([2, 1]))

    def test_format_prop_no_header(self):
        props = [('a', 'b'), ('cc', 'dd')]
        assert_that(UnityDoc.format_prop(props), equal_to(['a   b', 'cc  dd']))

    def test_format_prop_with_header(self):
        props = [('a', 'b')]
        assert_that(UnityDoc.format_prop(props, header=('name', 'value')),
                    equal_to(['name  value', 'a     b']))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.lib.common import is_valid
from storops.lib.converter import to_int_arr, to_disk_indices, to_int
from storops.vnx.resource.disk import VNXDiskList
from storops.vnx.resource.lun import VNXLun, VNXLunList
from storops.vnx.resource.snap import VNXSnap

__author__ = 'Cedric Zhuang'


def id_to_lun(lun_id):
    if is_valid(lun_id):
        ret = VNXLun(lun_id=to_int(lun_id))
    else:
        ret = None
    return ret


def name_to_lun(name):
    if is_valid(name):
        ret = VNXLun(name=name)
    else:
        ret = None
    return ret


def name_to_snap(name):
    if is_valid(name):
        ret = VNXSnap(name=name)
    else:
        ret = None
    return ret


def ids_to_lun_list(id_list_str):
    ids = to_int_arr(id_list_str)
    if ids:
        ret = VNXLunList(lun_ids=ids)
    else:
        ret = tuple()
    return ret


def indices_to_disk_list(disk_indices_str):
    disk_indices = to_disk_indices(disk_indices_str)
    if disk_indices:
        ret = VNXDiskList(disk_indices=disk_indices)
    else:
        ret = tuple()
    return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, instance_of, raises, none, not_none
from hamcrest import equal_to, has_length

from storops.exception import VNXMirrorLunNotAvailableError, \
    VNXMirrorNameInUseError, VNXMirrorAlreadyMirroredError, \
    VNXMirrorImageNotFoundError, VNXMirrorFractureImageError, \
    VNXMirrorSyncImageError, VNXMirrorPromoteNonLocalImageError, \
    VNXMirrorPromotePrimaryError, VNXMirrorFeatureNotAvailableError, \
    VNXMirrorNotFoundError, VNXDeleteMirrorWithSecondaryError, \
    VNXMirrorRemoveSynchronizingError, VNXMirrorGroupAlreadyMemberError, \
    VNXMirrorGroupMirrorNotMemberError, VNXMirrorGroupAlreadyPromotedError, \
    VNXMirrorGroupNameInUseError, VNXMirrorException
from storops_test.vnx.cli_mock import patch_cli
from storops_test.vnx.cli_mock import t_cli
from storops.vnx.enums import VNXMirrorViewRecoveryPolicy, \
    VNXMirrorViewSyncRate, VNXSPEnum, VNXMirrorImageState, \
    VNXMirrorGroupRecoveryPolicy
from storops.vnx.resource.mirror_view import VNXMirrorView, \
    VNXMirrorViewImage, VNXMirrorGroup, VNXMirrorGroupList, \
    VNXMirrorViewAsync, VNXMirrorGroupAsync, VNXMirrorGroupAsyncList

__author__ = 'Cedric Zhuang'


class VNXMirrorViewTest(TestCase):
    @patch_cli
    def test_get_all(self):
        mv_list = VNXMirrorView.get(t_cli())
        assert_that(len(mv_list), equal_to(4))

    @patch_cli(output='mirror_not_installed.txt')
    def test_mirror_view_not_installed(self):
        mv_list = VNXMirrorView.get(t_cli())
        assert_that(len(mv_list), equal_to(0))

        mv = VNXMirrorView.get(t_cli(), 'mv_sync_2')
        assert_that(mv.existed, equal_to(False))

    @patch_cli
    def test_get(self):
        mv = VNXMirrorView.get(t_cli(), 'mv_sync_2')
        assert_that(mv.uid, equal_to(
            '50:06:01:60:88:60:05:FE:04:00:00:00:00:00:00:00'))
        assert_that(mv.name, equal_to('mv_sync_2'))
        assert_that(mv.description, equal_to(''))
        assert_that(mv.logical_unit_numbers, 30)
        assert_that(mv.quiesce_threshold, equal_to(60))
        assert_that(mv.recovery_policy,
                    equal_to(VNXMirrorViewRecoveryPolicy.MANUAL))
        assert_that(len(mv.images), equal_to(2))
        assert_that(mv.images[0], instance_of(VNXMirrorViewImage))
        assert_that(mv.synchronization_rate,
                    equal_to(VNXMirrorViewSyncRate.MEDIUM))
        assert_that(mv.existed, equal_to(True))
        assert_that(mv.state, equal_to('Active'))
        assert_that(mv.image_transitioning, equal_to(False))
        assert_that(mv.image_size, equal_to(2097152))
        assert_that(mv.image_count, equal_to(2))
        assert_that(mv.image_faulted, equal_to(False))
        assert_that(mv.minimum_number_of_images_required, equal_to(0))
        assert_that(mv.write_intent_log_used, equal_to(True))
        assert_that(mv.synchronizing_progress, equal_to(100))
        assert_that(mv.remote_mirror_status, equal_to('Secondary Copy'))
        assert_that(mv.faulted, equal_to(False))
        assert_that(mv.transitioning, equal_to(False))
        assert_that(mv.is_primary, equal_to(False))

    @patch_cli
    def test_image_properties(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        assert_that(mv.is_primary, equal_to(True))
        assert_that(mv.primary_image.is_primary, equal_to(True))
        assert_that(mv.secondary_image.is_primary, equal_to(False))

    @patch_cli
    def test_create_success(self):
        mv = VNXMirrorView.create(t_cli(), 'mv0', 245)
        assert_that(mv.name, equal_to('mv0'))

    @patch_cli
    def test_create_lun_not_available_for_mirror(self):
        def f():
            VNXMirrorView.create(t_cli(), 'mv0', 244)

        assert_that(f, raises(VNXMirrorLunNotAvailableError, 'not available'))

    @patch_cli
    def test_create_name_in_use(self):
        def f():
            VNXMirrorView.create(t_cli(), 'mv0', 246)

        assert_that(f, raises(VNXMirrorNameInUseError, 'in use'))

    @patch_cli
    def test_add_image_success(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        mv.add_image('192.168.1.94', 71)
        assert_that(len(mv.images), equal_to(2))

    @patch_cli
    def test_add_image_already_mirrored(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.add_image('192.168.1.94', 72)

        assert_that(f, raises(VNXMirrorAlreadyMirroredError, 'exists'))

    @patch_cli
    def test_get_image_found(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        image = mv.get_image('50:06:01:60:88:60:05:FE')
        assert_that(image.state, equal_to(VNXMirrorImageState.SYNCHRONIZED))

    @patch_cli
    def test_get_image_not_found(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.get_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError, 'not found'))

    @patch_cli
    def test_remove_image_not_found(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.remove_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError, 'not found'))

    @patch_cli
    def test_remove_image_success(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        # no error raised
        mv.remove_image()

    @patch_cli
    def test_remove_image_no_secondary_image(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv1')
            mv.remove_image()

        assert_that(f,
                    raises(VNXMirrorImageNotFoundError, 'no secondary'))

    @patch_cli
    def test_fracture_primary_image(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.fracture_image('50:06:01:60:B6:E0:1C:F4')

        assert_that(f, raises(VNXMirrorFractureImageError, 'Cannot'))

    @patch_cli
    def test_fracture_image_success(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        # no error raised
        mv.fracture_image()

    @patch_cli
    def test_fracture_image_not_found(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.fracture_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError))

    @patch_cli
    def test_sync_image_not_found(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.sync_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError))

    @patch_cli
    def test_sync_image_failed(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.sync_image()

        assert_that(f, raises(VNXMirrorSyncImageError, 'failed'))

    @patch_cli
    def test_promote_image_not_found(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.promote_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError))

    @patch_cli
    def test_promote_non_local_image(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.promote_image()

        assert_that(f, raises(VNXMirrorPromoteNonLocalImageError,
                              'not local'))

    @patch_cli
    def test_promote_already_promoted(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv0')
            mv.promote_image('50:06:01:60:88:60:05:F0')

        assert_that(f, raises(VNXMirrorPromotePrimaryError, 'primary image'))

    @patch_cli
    def test_mirror_view_feature_not_installed(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv9')
            mv.delete()

        assert_that(f, raises(VNXMirrorFeatureNotAvailableError,
                              'not installed'))

    @patch_cli
    def test_delete_mirror_not_found_error(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv8')
            mv.delete()

        assert_that(f, raises(VNXMirrorNotFoundError, 'not found'))

    @patch_cli
    def test_delete_mirror_has_secondary(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv7')
            mv.delete()

        assert_that(f, raises(VNXDeleteMirrorWithSecondaryError,
                              'at least one secondary'))

    @patch_cli
    def test_remove_mirror_image_is_synchronizing(self):
        def f():
            mv = VNXMirrorView.get(t_cli(), 'mv2')
            mv.remove_image()

        assert_that(f, raises(VNXMirrorRemoveSynchronizingError,
                              'is being synchronized'))

    @patch_cli
    def test_force_delete_mirror_has_secondary(self):
        mv = VNXMirrorView.get(t_cli(), 'mv0')
        # no error raised
        mv.delete(force=True)


class VNXMirrorViewImageTest(TestCase):
    @patch_cli
    def test_properties(self):
        mv = VNXMirrorView.get(t_cli(), 'mv_sync_2')
        image = mv.get_image('50:06:01:60:88:60:05:FE')
        assert_that(image.uid, equal_to('50:06:01:60:88:60:05:FE'))
        assert_that(image.existed, equal_to(True))
        assert_that(image.is_primary, equal_to(True))
        assert_that(image.logical_unit_uid, equal_to(
            '60:06:01:60:41:C4:3D:00:6E:1C:50:9D:05:95:E5:11'))
        assert_that(image.condition, equal_to('Primary Image'))
        assert_that(image.state, none())
        assert_that(image.preferred_sp, equal_to(VNXSPEnum.SP_A))


class VNXMirrorGroupTest(TestCase):

    @patch_cli
    def test_create(self):
        mg = VNXMirrorGroup.create(t_cli(), name='test_group')
        assert_that(mg, instance_of(VNXMirrorGroup))

    @patch_cli
    def test_create_name_in_use(self):

        def _inner():
            VNXMirrorGroup.create(t_cli(), name='test_group_in_use')

        assert_that(_inner, raises(VNXMirrorGroupNameInUseError))

    @patch_cli
    def test_create_and_add(self):
        mirror = VNXMirrorView.get(t_cli(), name='mv_sync_2')
        mg = VNXMirrorGroup.create(t_cli(), name='petermg1', mirror=mirror)
        assert_that(mg, instance_of(VNXMirrorGroup))

    @patch_cli
    def test_get_single(self):
        mg = VNXMirrorGroup.get(t_cli(), name='petermg')
        assert_that(mg, instance_of(VNXMirrorGroup))
        assert_that(mg.name, equal_to('petermg'))
        assert_that(mg.gid, equal_to('50:06:01:60:B6:60:25:22:00:00:00:00'))
        assert_that(mg.description, equal_to(''))
        assert_that(mg.state, equal_to('Synchronized'))
        assert_that(mg.role, equal_to('Primary'))
        assert_that(mg.condition, equal_to('Active'))
        assert_that(mg.policy, equal_to(VNXMirrorGroupRecoveryPolicy.MANUAL))
        assert_that(mg.mirrors, has_length(2))
        assert_that(mg.group_mirrors, has_length(2))
        for m in mg.mirrors:
            assert_that(m, instance_of(VNXMirrorView))

        for mg in mg.group_mirrors:
            assert_that(
                mg.mirror_name,
                not_none())
            assert_that(mg.src_lun_id, instance_of(int))

    @patch_cli
    def test_get_all(self):
        mg_list = VNXMirrorGroup.get(t_cli())
        assert_that(len(mg_list), equal_to(2))
        assert_that(mg_list, instance_of(VNXMirrorGroupList))

    @patch_cli
    def test_promote_group(self):
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg1')
        mg1.promote_group()

    @patch_cli
    def test_fracture_group(self):
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg1')
        mg1.fracture_group()

    @patch_cli
    def test_add_to_group(self):
        mirror = VNXMirrorView.get(t_cli(), name='mv_sync_2')
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg1')
        mg1.add_mirror(mirror)

    @patch_cli
    def test_add_to_group_existed(self):
        mirror = VNXMirrorView.get(t_cli(), name='mv0')
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg')

        def _inner():
            mg1.add_mirror(mirror)

        assert_that(_inner, raises(VNXMirrorGroupAlreadyMemberError))

    @patch_cli
    def test_remove_from_group(self):
        mirror = VNXMirrorGroup.get(t_cli(), name='mv_sync_2')
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg')
        mg1.remove_mirror(mirror)

    @patch_cli
    def test_remove_from_group_already_removed(self):
        mirror = VNXMirrorGroup.get(t_cli(), name='not_in_group')
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg')

        def _inner():
            mg1.remove_mirror(mirror)

        assert_that(_inner, raises(VNXMirrorGroupMirrorNotMemberError))

    @patch_cli
    def test_sync_group(self):
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg')
        mg1.sync_group()

    @patch_cli
    def test_sync_group_already_promoted(self):
        mg1 = VNXMirrorGroup.get(t_cli(), name='mg_promote_on_primary')
        assert_that(mg1.sync_group, raises(VNXMirrorGroupAlreadyPromotedError))

    @patch_cli
    def test_delete_group(self):
        mg1 = VNXMirrorGroup.get(t_cli(), name='petermg')
        mg1.delete()


class VNXMirrorViewAsyncTest(TestCase):
    @patch_cli
    def test_get_all(self):
        mv_list = VNXMirrorViewAsync.get(t_cli())
        assert_that(len(mv_list), equal_to(2))

    @patch_cli
    def test_get(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_001')
        assert_that(mv.uid, equal_to(
            '8F:23:60:B6:60:01:06:50:08:00:00:00:00:00:00:00'))
        assert_that(mv.name, equal_to('testdr_001'))
        assert_that(mv.description, equal_to(''))
        assert_that(mv.logical_unit_numbers, 55)
        assert_that(mv.recovery_policy,
                    equal_to(VNXMirrorViewRecoveryPolicy.AUTO))
        assert_that(len(mv.images), equal_to(2))
        assert_that(mv.images[0], instance_of(VNXMirrorViewImage))
        assert_that(mv.synchronization_rate,
                    equal_to(VNXMirrorViewSyncRate.MEDIUM))
        assert_that(mv.existed, equal_to(True))
        assert_that(mv.state, equal_to('Active'))
        assert_that(mv.image_transitioning, equal_to(False))
        assert_that(mv.image_size, equal_to(104857600))
        assert_that(mv.image_count, equal_to(2))
        assert_that(mv.image_faulted, equal_to(False))
        assert_that(mv.minimum_number_of_images_required, equal_to(0))
        assert_that(mv.synchronizing_progress, equal_to(100))
        assert_that(mv.remote_mirror_status, equal_to('Mirrored'))
        assert_that(mv.faulted, equal_to(False))
        assert_that(mv.transitioning, equal_to(False))
        assert_that(mv.is_primary, equal_to(True))

    @patch_cli
    def test_image_properties(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_001')
        assert_that(mv.is_primary, equal_to(True))
        assert_that(mv.primary_image.is_primary, equal_to(True))
        assert_that(mv.secondary_image.is_primary, equal_to(False))

    @patch_cli
    def test_create_success(self):
        mv = VNXMirrorViewAsync.create(t_cli(), 'testdr_003', 71)
        assert_that(mv.name, equal_to('testdr_003'))

    @patch_cli
    def test_create_lun_not_available_for_mirror(self):
        def f():
            VNXMirrorViewAsync.create(t_cli(), 'mv0', 244)

        assert_that(f, raises(VNXMirrorException, 'LUN does not exist'))

    @patch_cli
    def test_create_name_in_use(self):
        def f():
            VNXMirrorViewAsync.create(t_cli(), 'testdr_003', 72)

        assert_that(f, raises(VNXMirrorNameInUseError, 'in use'))

    @patch_cli
    def test_add_image_success(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
        mv.add_image('192.168.1.94', 71)
        assert_that(len(mv.images), equal_to(2))

    @patch_cli
    def test_add_image_already_mirrored(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.add_image('192.168.1.94', 72)

        assert_that(f, raises(VNXMirrorAlreadyMirroredError, 'exists'))

    @patch_cli
    def test_get_image_found(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_004')
        image = mv.get_image('50:06:01:60:B6:60:23:7E')
        assert_that(image.state, equal_to(VNXMirrorImageState.SYNCHRONIZED))

    @patch_cli
    def test_get_image_not_found(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_004')
            mv.get_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorImageNotFoundError, 'not found'))

    @patch_cli
    def test_remove_image_not_found(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_004')
            mv.remove_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorException, 'image does not exist'))

    @patch_cli
    def test_remove_image_success(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_004')
        # no error raised
        mv.remove_image()

    @patch_cli
    def test_remove_image_no_secondary_image(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_003')
            mv.remove_image()

        assert_that(f,
                    raises(VNXMirrorImageNotFoundError, 'no secondary'))

    @patch_cli
    def test_fracture_primary_image(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.fracture_image('50:06:01:60:B6:60:23:8F')

        assert_that(f, raises(VNXMirrorException, 'does not exist'))

    @patch_cli
    def test_fracture_image_success(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
        # no error raised
        mv.fracture_image()

    @patch_cli
    def test_sync_image_not_found(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.sync_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorException, 'does not exist'))

    @patch_cli
    def test_sync_image_failed(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.sync_image()

        assert_that(f, raises(VNXMirrorException, 'already synchronized'))

    @patch_cli
    def test_promote_image_not_found(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.promote_image('50:06:01:60:88:60:05:FF')

        assert_that(f, raises(VNXMirrorException, 'does not exist'))

    @patch_cli
    def test_promote_non_local_image(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.promote_image()

        assert_that(f, raises(VNXMirrorException,
                              'promotion wasn\'t local'))

    @patch_cli
    def test_delete_mirror_not_found_error(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'mv8')
            mv.delete()

        assert_that(f, raises(VNXMirrorException, 'mirror does not exist'))

    @patch_cli
    def test_delete_mirror_has_secondary(self):
        def f():
            mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_005')
            mv.delete()

        assert_that(f, raises(VNXMirrorException,
                              'mirror with secondary images'))

    @patch_cli
    def test_force_delete_mirror_has_secondary(self):
        mv = VNXMirrorViewAsync.get(t_cli(), 'testdr_006')
        # no error raised
        mv.delete(force=True)


class VNXMirrorGroupAsyncTest(TestCase):

    @patch_cli
    def test_create(self):
        mg = VNXMirrorGroupAsync.create(t_cli(), name='test_group')
        assert_that(mg, instance_of(VNXMirrorGroupAsync))

    @patch_cli
    def test_create_name_in_use(self):

        def _inner():
            VNXMirrorGroupAsync.create(t_cli(), name='test_group_in_use')

        assert_that(_inner, raises(VNXMirrorException, 'same name as'))

    @patch_cli
    def test_create_and_add(self):
        mirror = VNXMirrorViewAsync.get(t_cli(), name='testdr_004')
        mg = VNXMirrorGroupAsync.create(t_cli(), name='petermg1',
                                        mirror=mirror)
        assert_that(mg, instance_of(VNXMirrorGroupAsync))

    @patch_cli
    def test_get_single(self):
        mg = VNXMirrorGroupAsync.get(t_cli(), name='petermg')
        assert_that(mg, instance_of(VNXMirrorGroupAsync))
        assert_that(mg.name, equal_to('petermg'))
        assert_that(mg.gid, equal_to('50:06:01:60:B6:60:23:8F:03:00:00:00'))
        assert_that(mg.description, equal_to(''))
        assert_that(mg.state, equal_to('Synchronized'))
        assert_that(mg.role, equal_to('Primary'))
        assert_that(mg.condition, equal_to('Normal'))
        assert_that(mg.policy, equal_to(VNXMirrorGroupRecoveryPolicy.AUTO))
        assert_that(len(mg.group_mirrors), equal_to(1))

    @patch_cli
    def test_get_all(self):
        mg_list = VNXMirrorGroupAsync.get(t_cli())
        assert_that(len(mg_list), equal_to(2))
        assert_that(mg_list, instance_of(VNXMirrorGroupAsyncList))

    @patch_cli
    def test_promote_group(self):
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg2')
        mg1.promote_group()

    @patch_cli
    def test_fracture_group(self):
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg1')
        mg1.fracture_group()

    @patch_cli
    def test_add_to_group(self):
        mirror = VNXMirrorViewAsync.get(t_cli(), name='testdr_004')
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg1')
        mg1.add_mirror(mirror)

    @patch_cli(output='mirror_-async_-addtogroup_-name_petermg1_'
                      '-mirrorname_testdr_004_ALREADYMEMBER.txt')
    def test_add_to_group_existed(self):
        mirror = VNXMirrorViewAsync.get(t_cli(), name='testdr_004')
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg1')

        def _inner():
            mg1.add_mirror(mirror)

        assert_that(_inner, raises(VNXMirrorGroupAlreadyMemberError))

    @patch_cli
    def test_remove_from_group(self):
        mirror = VNXMirrorGroupAsync.get(t_cli(), name='testdr_004')
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg1')
        mg1.remove_mirror(mirror)

    @patch_cli
    def test_remove_from_group_already_removed(self):
        mirror = VNXMirrorGroupAsync.get(t_cli(), name='testdr_003')
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg1')

        def _inner():
            mg1.remove_mirror(mirror)

        assert_that(_inner, raises(VNXMirrorGroupMirrorNotMemberError))

    @patch_cli
    def test_sync_group(self):
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg')
        mg1.sync_group()

    @patch_cli
    def test_delete_non_empty_group(self):
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='petermg')

        def _inner():
            mg1.delete()
        assert_that(_inner, raises(VNXMirrorException, 'still has members'))

    @patch_cli
    def test_delete_group(self):
        mg1 = VNXMirrorGroupAsync.get(t_cli(), name='test_group')
        mg1.delete()
